/*
 * Copyright (c) 2008 NVIDIA Corporation.  All rights reserved.
 *
 * NVIDIA Corporation and its licensors retain all intellectual property
 * and proprietary rights in and to this software, related documentation
 * and any modifications thereto.  Any use, reproduction, disclosure or
 * distribution of this software and related documentation without an express
 * license agreement from NVIDIA Corporation is strictly prohibited.
 */

#include "sensor_common.h"
#include "nvodm_imager.h"
#include "nvodm_imager_int.h"
#include "imager_hal.h" //shard ressources with driver
#include "I2clib.h"
#include "GPIOLib.h"

#include "sensor_bayer_camera_config_min.h"

/****************************************************************************
 * Implementing for a New Sensor:
 * 
 * When implementing for a new sensor from this template, there are four 
 * phases:
 * Phase 0: Gather sensor information. In phase 0, we should gather the sensor
 *          information for bringup such as sensor capabilities, set mode i2c
 *          write sequence, NvOdmIoAddress in nvodm_query_discovery_imager.h,
 *          and so on.
 * Phase 1: Bring up a sensor. After phase 1, we should be able to do a basic
 *          still capture with minimal interactions between camera driver
 *          and the sensor
 * Phase 2: Calibrate a sensor. After phase 2, we should be able to calibrate
 *          a sensor. Calibration will need some interactions between camera
 *          driver and the sensor such as setting exposure, gains, and so on.
 * Phase 3: Fully functioning sensor. After phase 3, the sensor should be fully
 *          functioning as described in nvodm_imager.h
 *
 * To help implementing for a new sensor, the template code will be marked as
 * Phase 0, 1, 2, or 3. 
 ****************************************************************************/

//For I2C
#define SEQUENCE_WAIT_MS                (0xFFFE)
#define SEQUENCE_END                    (0xFFFF)
#define SEQUENCE_FAST_SETMODE_START     (0xFFFD)
#define SEQUENCE_FAST_SETMODE_END       (0xFFFC)


/**
 * Sensor default settings. Phase 2. Sensor Dependent.
 * #defines make the code more readable and helps maintainability
 * The template tries to keep the list reduced to just things that
 * are used more than once.  Ideally, all numbers in the code would use
 * a #define.
 */
#define SENSOR_BAYER_DEFAULT_EXPOSURE_FINE      (0x0000) //GER: Compute correct value
#define SENSOR_BAYER_DEFAULT_PLL_MULT           (0x0001) //DAN
#define SENSOR_BAYER_DEFAULT_PLL_PRE_DIV        (0x0001) //DAN
#define SENSOR_BAYER_DEFAULT_PLL_VT_PIX_DIV     (0x0001) //DAN
#define SENSOR_BAYER_DEFAULT_PLL_VT_SYS_DIV     (0x0001)
#define SENSOR_BAYER_DEFAULT_MAX_GAIN           (63)
#define SENSOR_BAYER_DEFAULT_MIN_GAIN           (8)
#define SENSOR_BAYER_DEFAULT_GAIN_R             (8)
#define SENSOR_BAYER_DEFAULT_GAIN_GR            (8)
#define SENSOR_BAYER_DEFAULT_GAIN_GB            (8)
#define SENSOR_BAYER_DEFAULT_GAIN_B             (8)
#define SENSOR_BAYER_DEFAULT_MAX_FRAME_LENGTH   (0xFFFF)
#define SENSOR_BAYER_DEFAULT_MIN_FRAME_LENGTH   (1976+21)

#define SENSOR_BAYER_DEFAULT_LINE_LENGTH_FULL       (2*(2592/2+346+64+80/2)) // 2*(W/2+HBmin)
#define SENSOR_BAYER_DEFAULT_FRAME_LENGTH_FULL      (1944+9)
#define SENSOR_BAYER_DEFAULT_EXPOSURE_COARSE_FULL   (1943)

#define SENSOR_BAYER_DEFAULT_LINE_LENGTH_QUAR       (0x0910)
#define SENSOR_BAYER_DEFAULT_FRAME_LENGTH_QUAR      (0x03E8)
#define SENSOR_BAYER_DEFAULT_EXPOSURE_COARSE_QUAR   (0x03E8)

#define SENSOR_BAYER_DEFAULT_LINE_LENGTH_VGA_BIN       (2*(640/2+346*4+64+20/2)) // 2*(W/2+HBmin) ; HBmin=346*(row_bin+1)+64+(WDC/2)
#define SENSOR_BAYER_DEFAULT_FRAME_LENGTH_VGA_BIN      (480+9)
#define SENSOR_BAYER_DEFAULT_EXPOSURE_COARSE_VGA_BIN   (479)

#define SENSOR_GAIN_TO_F32(x)     ((NvF32)(x)/8.0f)//(128.f / (128.f - (NvF32)(x)))
#define SENSOR_F32_TO_GAIN(x)     ((NvU32)((x)*8.0f))//((NvU32)(128.f - (128.f / (x))))


#define SODIMM_PIN_SENSOR_ENABLE	(176) //(77)	//Low active 
#define SODIMM_PIN_SENSOR_RESET		(174) //(69)	//Low active
//#define I2C_ADDRESS					(0x90>>2) //(0x30)
#define I2C_ADDRESS					(0x3C>>2) //(0x30)

/**
 * ----------------------------------------------------------
 *  Start of Phase 0 code
 * ----------------------------------------------------------
 */

/**
 * Sensor Specific Variables/Defines. Phase 0.
 *
 * If a new sensor code is created from this template, the following
 * variable/defines should be created for the new sensor specifically.
 * 1. A sensor capabilities (NvOdmImagerCapabilities) for this specific sensor
 * 2. A sensor set mode sequence list  (an array of SensorSetModeSequence)
 *    consisting of pairs of a sensor mode and corresponding set mode sequence
 * 3. A set mode sequence for each mode defined in 2.
 */

/**
 * Sensor bayer's capabilities. Phase 0. Sensor Dependent.
 * This is the first thing the camera driver requests. The information here
 * is used to setup the proper interface code (CSI, VIP, pixel type), start
 * up the clocks at the proper speeds, and so forth.
 * For Phase 0, one could start with these below as an example for a bayer
 * sensor and just update clocks from what is given in the reference documents
 * for that device. The FAE may have specified the clocks when giving the
 * reccommended i2c settings.
 */
static NvOdmImagerCapabilities g_SensorBayerCaps =
{
    "MT9P031",  // string identifier, used for debugging
    NvOdmImagerSensorInterface_Parallel8, // interconnect type
    {
    NvOdmImagerPixelType_BayerRGGB,
    //NvOdmImagerPixelType_BayerBGGR,
    //NvOdmImagerPixelType_BayerGRBG,
    //NvOdmImagerPixelType_BayerGBRG,
    //NvOdmImagerPixelType_UYVY,
	},
    NvOdmImagerOrientation_0_Degrees, // rotation with respect to main display
    NvOdmImagerDirection_Toward,          // toward or away the main display
    6000, // Initial sensor clock (in kHz).
    { {48000, 3.0}, {0,0} }, // preferred clock profile
    {
        NvOdmImagerSyncEdge_Rising, // hsync polarity
        NvOdmImagerSyncEdge_Rising, // vsync polarity
        NV_TRUE,                   // uses mclk on vgp0
    }, // parallel settings (VIP)
    { 0 }, // serial settings (CSI) (not used)
    { 0, 0 }, // min blank time, shouldn't need adjustment
    0, // preferred mode, which resolution to use on startup
    0, // FOCUSER_GUID, only set if focuser code available
    0, // FLASH_GUID, only set if flash device code is available
    NVODM_IMAGER_CAPABILITIES_END
};

 
/*
static DWORD WriteToI2C(DevCtrlReg16* SetModeSequence)
{
	char data[10];
	
	DevCtrlReg16 *ptr = SetModeSequence;
	DWORD ret= TRUE;

	if (I2CInit())
	{
		if(GetI2CLock(500)){
			//I2CSetSpeed(I2C_100KBPS);				// Currently, this sets LastError=1210 on Tegra
			
			while(ptr->RegAddr!= SEQUENCE_END)
			{
				if(ptr->RegAddr == SEQUENCE_WAIT_MS){
					Sleep(1000*ptr->RegValue);
				}
				else{
					//switch bytes
					data[1] = (char)ptr->RegValue;
					data[0] = (char)(ptr->RegValue>>8);
					if(!I2CBurstWrite(I2C_ADDRESS, data, (char)ptr->RegAddr, 2))
					{
						NKDbgPrintfW(L"WriteToI2C I2CBurstWrite faild\n\r");
						ret= FALSE;
						break;
					}
				}
				ptr += 1;
			}
			ReleaseI2CLock();
		}
		else{
			ret= FALSE;
			NKDbgPrintfW(L"WriteToI2C GetI2CLock faild\n\r");
		}
		I2CDeInit();
	}
	else{
		ret= FALSE;
		NKDbgPrintfW(L"WriteToI2C I2CInit faild\n\r");
	}
	return ret;
}
*/

enum Registers_5MP
{
    CHIP_VERSION           = 0x00, 
    ROW_START              = 0x01, 
    COLUMN_START           = 0x02, 
    ROW_SIZE               = 0x03, 
    COLUMN_SIZE            = 0x04, 
    HORIZONTAL_BLANK       = 0x05, 
    VERTICAL_BLANK         = 0x06, 
    OUTPUT_CONTROL         = 0x07, 
    SHUTTER_WIDTH_UPPER    = 0x08, 
    SHUTTER_WIDTH_LOWER    = 0x09, 
    PIXEL_CLOCK_CONTROL    = 0x0A, 
    RESTART                = 0x0B, 
    SHUTTER_DELAY          = 0x0C, 
    RESET                  = 0x0D, 
    PLL_CONTROL            = 0x10, 
    PLL_CONFIG_1           = 0x11, 
    PLL_CONFIG_2           = 0x12, 
    READ_MODE_1            = 0x1E, 
    READ_MODE_2            = 0x20, 
    ROW_ADDRESS_MODE       = 0x22, 
    COLUMN_ADDRESS_MODE    = 0x23, 
    GREEN1_GAIN            = 0x2B, 
    BLUE_GAIN              = 0x2C, 
    RED_GAIN               = 0x2D, 
    GREEN2_GAIN            = 0x2E, 
    GLOBAL_GAIN            = 0x35, 
    TEST_PATTERN_CONTROL   = 0xA0, 
    TEST_PATTERN_GREEN     = 0xA1, 
    TEST_PATTERN_RED       = 0xA2, 
    TEST_PATTERN_BLUE      = 0xA3, 
    TEST_PATTERN_BAR_WIDTH = 0xA4, 
    CHIP_VERSION_ALT       = 0xFF  
}; 

/**
 * SetMode Sequence for 2592x1944. Phase 0. Sensor Dependent.
 * This sequence should put sensor in streaming mode for 2592x1944
 * This is usually given by the FAE or the sensor vendor.
 */
//
//static DevCtrlReg16 SetModeSequence_640x480[] =
//{
//	{READ_MODE_1, 0x4406 },
//	{ROW_SIZE, 480-1 }, 
//	{ROW_START, 8 + 500 },
//	{COLUMN_START,16 + 500},
//	{COLUMN_SIZE, 640-1},
//	{SHUTTER_WIDTH_LOWER, 300},
//	//{0x0103, VERTICAL_BLANK},
//	{SEQUENCE_END,0}
//};

static DevCtrlReg16 SetModeSequence_640x480[] =
{
#if 1
	{READ_MODE_1, 0x4406 },
	//{READ_MODE_2, 0x0000 }, //Disable Row BLC
	{COLUMN_SIZE, 2559 },
	{ROW_SIZE, 1919 },
	{SHUTTER_WIDTH_LOWER, 450},
	{ROW_ADDRESS_MODE, ((4-1)<<4)|(4-1)}, //vertical resolution/4 with binning 
	{COLUMN_ADDRESS_MODE, ((4-1)<<4)|(4-1)}, //horizontal resolution/4 with binning
	//{PIXEL_CLOCK_CONTROL,(1<<15)},
	//{GLOBAL_GAIN,0},
	//{GREEN1_GAIN,0},
	//{RED_GAIN,0},
	//{BLUE_GAIN,30},
	//{GREEN2_GAIN,0},
	//{TEST_PATTERN_CONTROL,(4<<3)},
	//{TEST_PATTERN_GREEN, 2000},
	//{TEST_PATTERN_RED, 2000},
	//{TEST_PATTERN_BLUE, 2000},
	//{TEST_PATTERN_BAR_WIDTH, 201},
	//{TEST_PATTERN_CONTROL,(4<<3)|1},
	//{PIXEL_CLOCK_CONTROL,(1<<15) },	//invert pixel clock
	//{ROW_SIZE, 480-1 }, 
	//{ROW_START, 8 + 4+500 },
	//{COLUMN_START,16 + 4+500},
	//{COLUMN_SIZE, 640-1},
	//{SHUTTER_WIDTH_LOWER, 300},
	////{0x0103, VERTICAL_BLANK},
	{SEQUENCE_END,0}
#else
	{ READ_MODE_1,				 0x4406 },
	{ SHUTTER_WIDTH_LOWER,			450 },
	{ VERTICAL_BLANK,				 23 },
	{ HORIZONTAL_BLANK,				100 },

	{ COLUMN_SIZE,					639 },
	{ ROW_SIZE,						479 },

	{ ROW_ADDRESS_MODE,			      0 }, 
	{ COLUMN_ADDRESS_MODE,            0 },
	{ ROW_START,            8 + 4 + 500 },
	{ COLUMN_START,        16 + 4 + 500 },
	{ SEQUENCE_END,                   0 }
#endif
};

static DevCtrlReg16 SetModeSequence_640x480_TestMode[] =
{

	{READ_MODE_1, 0x4406 },
	//{ROW_SIZE, 480-1 }, 
	//{ROW_START, 8 + 4 },
	//{COLUMN_START,16 + 4},
	//{COLUMN_SIZE, 640-1},
	{SHUTTER_WIDTH_LOWER, 300},
	{ROW_ADDRESS_MODE, ((4-1)<<4)|(4-1)}, //vertical resolution/4 with binning 
	{COLUMN_ADDRESS_MODE, ((4-1)<<4)|(4-1)}, //horizontal resolution/4 with binning
	{TEST_PATTERN_CONTROL,(6<<3)},
	{TEST_PATTERN_GREEN, 2000},
	{TEST_PATTERN_RED, 2000},
	{TEST_PATTERN_BLUE, 2000},
	{TEST_PATTERN_BAR_WIDTH, 201},
	{TEST_PATTERN_CONTROL,(6<<3)|1},
	{SEQUENCE_END,0}
};
//
static DevCtrlReg16 SetModeSequence_2592x1944[] =
{
    {READ_MODE_1, 0x4406 },
    {SEQUENCE_END,0}
};



typedef struct ModeDependentSettingsRec
{
    NvU32 LineLength;
    NvU32 FrameLength;
    NvU32 MinFrameLength;
    NvU32 CoarseTime;
} ModeDependentSettings;

static ModeDependentSettings ModeDependentSettings_1316x988 = 
{
    SENSOR_BAYER_DEFAULT_LINE_LENGTH_QUAR,
    SENSOR_BAYER_DEFAULT_FRAME_LENGTH_QUAR,
    SENSOR_BAYER_DEFAULT_FRAME_LENGTH_QUAR,
    SENSOR_BAYER_DEFAULT_EXPOSURE_COARSE_QUAR,
};

static ModeDependentSettings ModeDependentSettings_2632x1976 = 
{
    SENSOR_BAYER_DEFAULT_LINE_LENGTH_FULL,
    SENSOR_BAYER_DEFAULT_FRAME_LENGTH_FULL,
    SENSOR_BAYER_DEFAULT_FRAME_LENGTH_FULL,
    SENSOR_BAYER_DEFAULT_EXPOSURE_COARSE_FULL,
};

static ModeDependentSettings ModeDependentSettings_640x480 = 
{
    SENSOR_BAYER_DEFAULT_LINE_LENGTH_VGA_BIN,
    SENSOR_BAYER_DEFAULT_FRAME_LENGTH_VGA_BIN,
    SENSOR_BAYER_DEFAULT_FRAME_LENGTH_VGA_BIN,
    SENSOR_BAYER_DEFAULT_EXPOSURE_COARSE_VGA_BIN,
};

/**
 * A list of sensor mode and its SetModeSequence. Phase 0. Sensor Dependent.
 * When SetMode is called, it searches the list for desired resolution and
 * frame rate and write the SetModeSequence through i2c.
 * For Phase 3, more resolutions will be added to this table, allowing
 * for peak performance during Preview, D1 and 720P encode.
 */
static SensorSetModeSequence g_SensorBayerSetModeSequenceList[] =
{
     // WxH         x,y    fps    set mode sequence
    //{{Width x Height},			 {x_skip, y_skip}, fps,  AR, PLL Mult}, Set Mode Sequence
   {{{640, 480}, {0, 0}, 26.5, 1.0, 3.0}, SetModeSequence_640x480,&ModeDependentSettings_640x480},
   //{{{640, 480}, {0, 0}, 30, 1.0, 3.0}, SetModeSequence_640x480_TestMode,&ModeDependentSettings_640x480},
//    {{{2632, 1976}, {2, 2}, 1.5,}, SetModeSequence_2632x1976,&ModeDependentSettings_2632x1976},
   {{{2592, 1944}, {0, 0}, 7, 1.0, 3.0}, SetModeSequence_2592x1944,&ModeDependentSettings_2632x1976}
};


/**
 * ----------------------------------------------------------
 *  Start of Phase 1, Phase 2, and Phase 3 code
 * ----------------------------------------------------------
 */

/**
 * Sensor bayer's private context. Phase 1.
 */
typedef struct SensorBayerContextRec
{
    // Phase 1 variables.
    //NvOdmImagerI2cConnection I2c;
    NvU32 ModeIndex;
    NvU32 NumModes;
    NvOdmImagerPowerLevel PowerLevel;


    // Phase 2 variables.
    NvBool SensorInitialized;

    NvU32 SensorInputClockkHz; // mclk (extclk)
    NvF32 SensorClockRatio; // pclk/mclk

    NvF32 Exposure;
    NvF32 MaxExposure;
    NvF32 MinExposure;

    NvF32 Gains[4];
    NvF32 MaxGain;
    NvF32 MinGain;

    NvF32 FrameRate;
    NvF32 RequestedMaxFrameRate; // requested limit
    NvF32 MinFrameRate; // sensor's limit
    NvF32 MaxFrameRate; // sensor's limit

    // Phase 2 variables. Sensor Dependent.
    // These are used to set or get exposure, frame rate, and so on.
	
    NvU32 PllMult;
    NvU32 PllPreDiv;
    NvU32 PllVtPixDiv;
    NvU32 PllVtSysDiv;

    NvU32 CoarseTime;
    NvU32 VtPixClkFreqHz;

    NvU32 LineLength;
    NvU32 FrameLength;

    NvU32 MaxFrameLength;
    NvU32 MinFrameLength;

    // Phase 3.
    NvBool FastSetMode;


}SensorBayerContext;


/**
 * Static functions
 * Some compilers require these to be declared upfront.
 */
static NvBool SensorBayer_Open(NvOdmImagerHandle hImager);

static void SensorBayer_Close(NvOdmImagerHandle hImager);

static void
SensorBayer_GetCapabilities(
    NvOdmImagerHandle hImager,
    NvOdmImagerCapabilities *pCapabilities);

static void
SensorBayer_ListModes(
    NvOdmImagerHandle hImager,
    NvOdmImagerSensorMode *pModes,
    NvS32 *pNumberOfModes);

static NvBool
SensorBayer_SetMode(
    NvOdmImagerHandle hImager,
    const SetModeParameters *pParameters,
    NvOdmImagerSensorMode *pSelectedMode,
    SetModeParameters *pResult);

static NvBool
SensorBayer_SetPowerLevel(
    NvOdmImagerHandle hImager,
    NvOdmImagerPowerLevel PowerLevel);

static NvBool
SensorBayer_SetParameter(
    NvOdmImagerHandle hImager,
    NvOdmImagerParameter Param,
    NvS32 SizeOfValue,
    const void *pValue);

static NvBool
SensorBayer_GetParameter(
    NvOdmImagerHandle hImager,
    NvOdmImagerParameter Param,
    NvS32 SizeOfValue,
    void *pValue);

static void
SensorBayer_GetPowerLevel(
    NvOdmImagerHandle hImager,
    NvOdmImagerPowerLevel *pPowerLevel);

static NvBool
SensorBayer_WriteFramerate(
        SensorBayerContext *pContext,
        NvF32 *pFrameRate,
        NvBool WriteNewFrameRate);

/**
 * SensorBayer_WriteExposure. Phase 2. Sensor Dependent.
 * Calculate and write the values of the sensor registers for the new
 * exposure time.
 * Without this, calibration will not be able to capture images
 * at various brightness levels, and the final product won't be able
 * to compensate for bright or dark scenes. 
 */
static NvBool
SensorBayer_WriteExposure(
    SensorBayerContext *pContext,
    const NvF32 *pNewExposureTime)
{

	//NKDbgPrintfW(L"SensorBayer_WriteExposure NOT IMPLEMENTED\n\r");
    NvF32 Freq = (NvF32)pContext->VtPixClkFreqHz;
    NvF32 LineLength = (NvF32)pContext->LineLength;
    NvF32 ExpTime = *pNewExposureTime;
    NvU32 NewCoarseTime = 0;
    NvU32 NewFrameLength = 0;


    if (*pNewExposureTime > pContext->MaxExposure ||
        *pNewExposureTime < pContext->MinExposure)
        return NV_FALSE;

    //// Here, we have to decide if the new exposure time is valid
    //// based on the sensor and current sensor settings.
    NewCoarseTime = (NvU32)((Freq * ExpTime) / LineLength);
    NewFrameLength = NewCoarseTime + 1;

    //// Consider requested max frame rate,
    //if (pContext->RequestedMaxFrameRate > 0.0f)
    //{
    //    NvU32 RequestedMinFrameLengthLines = 0;
    //    RequestedMinFrameLengthLines =
    //       (NvU32)(Freq / (LineLength * pContext->RequestedMaxFrameRate));

    //    if (NewFrameLength < RequestedMinFrameLengthLines)
    //        NewFrameLength = RequestedMinFrameLengthLines;
    //}

    //// Clamp to sensor's limit
    //if (NewFrameLength > pContext->MaxFrameLength)
    //    NewFrameLength = pContext->MaxFrameLength;
    //else if (NewFrameLength < pContext->MinFrameLength)
    //    NewFrameLength = pContext->MinFrameLength;
    //
    //// Need to update FrameLength?
    //if (NewFrameLength != pContext->FrameLength)
    //{
    //    NVODM_WRITE_RETURN_ON_ERROR(&pContext->I2c,
    //            0x0340, (NewFrameLength >> 8) & 0xFF);
    //    NVODM_WRITE_RETURN_ON_ERROR(&pContext->I2c,
    //            0x0341, NewFrameLength & 0xFF);
    //    pContext->FrameLength = NewFrameLength;

    //    // Update variables depending on FrameLength.
    //    pContext->FrameRate = (NvF32)pContext->VtPixClkFreqHz /
    //                          (NvF32)(pContext->FrameLength *
    //                                  pContext->LineLength);
    //}
    //
    //// FrameLength should have been updated but we have to check again
    //// in case FrameLength gets clamped.
    //if (NewCoarseTime >= pContext->FrameLength)
    //{
    //    NewCoarseTime = pContext->FrameLength - 1;
    //}
    //
    //// Current CoarseTime needs to be updated?
    if (pContext->CoarseTime != NewCoarseTime)
    {
    //    NVODM_WRITE_RETURN_ON_ERROR(&pContext->I2c,
    //            0x0202, (NewCoarseTime >> 8) & 0xFF);
    //    NVODM_WRITE_RETURN_ON_ERROR(&pContext->I2c,
    //            0x0203, NewCoarseTime & 0xFF);
        SensorI2C_Open();
        SensorI2C_Write(SHUTTER_WIDTH_LOWER, NewCoarseTime);
        SensorI2C_Close();

        // Calculate the new exposure based on the sensor and sensor
        // settings.
        pContext->Exposure = ((NvF32)NewCoarseTime * LineLength +
                              SENSOR_BAYER_DEFAULT_EXPOSURE_FINE) /
                             Freq;
        pContext->CoarseTime = NewCoarseTime;
    }

    return NV_TRUE;
}

/**
 * SensorBayer_SetInputClock. Phase 2. Sensor Dependent.
 * Setting the input clock and updating the variables based on the input clock
 * The frame rate and exposure calculations are directly related to the
 * clock speed, so this is how the camera driver (the controller of the clocks)
 * can inform the nvodm driver of the current clock speed.
 */
static NvBool
SensorBayer_SetInputClock(
    SensorBayerContext *pContext,
    const NvOdmImagerClockProfile *pClockProfile)
{

    if (pClockProfile->ExternalClockKHz <
        g_SensorBayerCaps.InitialSensorClockRateKHz)
        return NV_FALSE;

    pContext->SensorInputClockkHz = pClockProfile->ExternalClockKHz;
    pContext->SensorClockRatio = pClockProfile->ClockMultiplier;

    // if sensor is initialized (SetMode has been called), then we need to
    // update necessary sensor variables (based on this particular sensor)
    if (pContext->SensorInitialized)
    {
        pContext->VtPixClkFreqHz =
            (NvU32)((pContext->SensorInputClockkHz * 1000 * pContext->PllMult) /
                    (pContext->PllPreDiv * pContext->PllVtPixDiv *
                     pContext->PllVtSysDiv));

        pContext->Exposure = (NvF32)(pContext->CoarseTime * pContext->LineLength +
                                     SENSOR_BAYER_DEFAULT_EXPOSURE_FINE) /
                             (NvF32)pContext->VtPixClkFreqHz;
        pContext->MaxExposure = (NvF32)((pContext->FrameLength - 1) * pContext->LineLength +
                                        SENSOR_BAYER_DEFAULT_EXPOSURE_FINE) /
                                (NvF32)pContext->VtPixClkFreqHz;
        pContext->MinExposure = (NvF32)(1 * pContext->LineLength +
                                    SENSOR_BAYER_DEFAULT_EXPOSURE_FINE) /
                                (NvF32)pContext->VtPixClkFreqHz;

    }

    return NV_TRUE;
}


/** 
 * SensorBayer_WriteGains. Phase 2. Sensor Dependent.
 * Writing the sensor registers for the new gains.
 * Just like exposure, this allows the sensor to brighten an image. If the
 * exposure time is insufficient to make the picture viewable, gains are 
 * applied.  During calibration, the gains will be measured for maximum 
 * effectiveness before the noise level becomes too high.
 * If per-channel gains are available, they are used to normalize the color.
 * Most sensors output overly greenish images, so the red and blue channels
 * are increased.
 */
static NvBool
SensorBayer_WriteGains(
    SensorBayerContext *pContext,
    const NvF32 *pGains)
{
	//NKDbgPrintfW(L"SensorBayer_WriteGains NOT IMPLEMENTED\n\r");
	DevCtrlReg16 GainSequence[] =
	{
		{GREEN1_GAIN,0},
		{RED_GAIN,0},
		{BLUE_GAIN,0},
		{GREEN2_GAIN,0},
        {SEQUENCE_END, 0}
	};
    NvU32 i = 0;
    NvU16 NewGains[4] = {0};
	

    for(i = 0; i < 4; i++)
    {
        if (pGains[i] > pContext->MaxGain)
            return NV_FALSE;

        // convert float to gain's unit
        GainSequence[i].RegValue = (NvU16)SENSOR_F32_TO_GAIN(pGains[i]);
    }

	SensorI2C_WriteSequence(GainSequence);
    memcpy(pContext->Gains, pGains, sizeof(NvF32)*4);

    return NV_TRUE;
}


/**
 * SensorBayer_Open. Phase 1.
 * Initialize sensor bayer and its private context
 */
static NvBool SensorBayer_Open(NvOdmImagerHandle hImager)
{

    SensorBayerContext *pSensorBayerContext = NULL;

    if (!hImager || !hImager->pSensor)
        return NV_FALSE;

    pSensorBayerContext =
        (SensorBayerContext*)malloc(sizeof(SensorBayerContext));
    if (!pSensorBayerContext)
        goto fail;

    memset(pSensorBayerContext, 0, sizeof(SensorBayerContext));

    DEBUGMSG(ZONE_FUNC, (_T("+SensorYuv_Open\n")));
  
	//Set Multiplexed GPIOs
	SetGPIOAltFn(TEGRA_GPIONUM('b',4),-1,DIR_IN); //CIF_D7
	SetGPIOAltFn(TEGRA_GPIONUM('a',7),-1,DIR_IN); //CIF_D6


	//Do Hardreset and Enable Sensor
	SetPinAltFn(SODIMM_PIN_SENSOR_ENABLE,-1,DIR_OUT);
	SetPinAltFn(SODIMM_PIN_SENSOR_RESET,-1,DIR_OUT);
	SetPinLevel(SODIMM_PIN_SENSOR_ENABLE,0);
	SetPinLevel(SODIMM_PIN_SENSOR_RESET,0);
	Sleep(50);
	SetPinLevel(SODIMM_PIN_SENSOR_RESET,1);	//release reset




    pSensorBayerContext->NumModes =
        NV_ARRAY_SIZE(g_SensorBayerSetModeSequenceList);
    pSensorBayerContext->ModeIndex =
        pSensorBayerContext->NumModes + 1; // invalid mode
    
    /** 
     * Phase 2. Initialize exposure and gains.
     */
    pSensorBayerContext->Exposure = -1.0; // invalid exposure
    //pSensorBayerContext->MinExposure = 1.0;
    //pSensorBayerContext->MaxExposure = 1.0;

     // convert gain's unit to float
    pSensorBayerContext->MaxGain = SENSOR_GAIN_TO_F32(SENSOR_BAYER_DEFAULT_MAX_GAIN);
    pSensorBayerContext->MinGain = SENSOR_GAIN_TO_F32(SENSOR_BAYER_DEFAULT_MIN_GAIN);

    pSensorBayerContext->MinFrameRate = 0;
    pSensorBayerContext->MaxFrameRate = 120;
    
    /**
     * Phase 2 ends here.
     */

    hImager->pSensor->pPrivateContext = pSensorBayerContext;

    g_SensorI2C_Addr = I2C_ADDRESS;
    g_SensorI2C_Speed = I2C_100KBPS;
    g_SensorI2C_OffsetWidth = 1;
    g_SensorI2C_DataWidth = 2;

    DEBUGMSG(ZONE_FUNC, (_T("-SensorYuv_Open\n")));

    return NV_TRUE;    

fail:
    free(pSensorBayerContext);
    return NV_FALSE;
}


/**
 * SensorBayer_Close. Phase 1.
 * Free the sensor's private context
 */
static void SensorBayer_Close(NvOdmImagerHandle hImager)
{
    SensorBayerContext *pContext = NULL;
        
    if (!hImager || !hImager->pSensor || !hImager->pSensor->pPrivateContext)
        return;

    pContext = (SensorBayerContext*)hImager->pSensor->pPrivateContext;

    DEBUGMSG(ZONE_FUNC, (_T("+SensorYuv_Close\n")));

    // cleanup
    free(pContext);
    hImager->pSensor->pPrivateContext = NULL;

    DEBUGMSG(ZONE_FUNC, (_T("+SensorYuv_Close\n")));
}

/**
 * SensorBayer_GetCapabilities. Phase 1.
 * Returnt sensor bayer's capabilities
 */
static void
SensorBayer_GetCapabilities(
    NvOdmImagerHandle hImager,
    NvOdmImagerCapabilities *pCapabilities)
{
    DEBUGMSG(ZONE_FUNC, (_T("+SensorYuv_GetCapabilities\n")));
    // Copy the sensor caps from g_SensorBayerCaps
    memcpy(pCapabilities, &g_SensorBayerCaps,
        sizeof(NvOdmImagerCapabilities));
    DEBUGMSG(ZONE_FUNC, (_T("-SensorYuv_GetCapabilities\n")));
}

/**
 * SensorBayer_ListModes. Phase 1.
 * Return a list of modes that sensor bayer supports.
 * If called with a NULL ptr to pModes, then it just returns the count
 */
static void
SensorBayer_ListModes(
    NvOdmImagerHandle hImager,
    NvOdmImagerSensorMode *pModes,
    NvS32 *pNumberOfModes)
{
    SensorBayerContext *pContext =
        (SensorBayerContext*)hImager->pSensor->pPrivateContext;
    NvS32 i;

    DEBUGMSG(ZONE_FUNC, (_T("+SensorYuv_ListModes\n")));

    if (pNumberOfModes)
    {
        *pNumberOfModes = (NvS32)pContext->NumModes;
        if (pModes)
        {
            // Copy the modes from g_SensorBayerSetModeSequenceList
            for (i = 0; i < *pNumberOfModes; i++)
                pModes[i] = g_SensorBayerSetModeSequenceList[i].Mode;
        }
    }
    DEBUGMSG(ZONE_FUNC, (_T("-SensorYuv_ListModes\n")));
}

static NV_INLINE NvBool SensorBayerChooseModeIndex(
    SensorBayerContext *pContext,
    NvSize Resolution,
    NvU32 *pIndex)
{
    NvU32 Index;

    for (Index = 0; Index < pContext->NumModes; Index++)
    {
        if ((Resolution.width ==
             g_SensorBayerSetModeSequenceList[Index].Mode.
             ActiveDimensions.width) &&
            (Resolution.height ==
             g_SensorBayerSetModeSequenceList[Index].Mode.
             ActiveDimensions.height))
        {
            *pIndex = Index;
            return NV_TRUE;
        }
    }

    return NV_FALSE;
}

/**
 * SensorBayer_SetMode. Phase 1.
 * Set sensor bayer to the mode of the desired resolution and frame rate.
 */
static NvBool
SensorBayer_SetMode(
    NvOdmImagerHandle hImager,
    const SetModeParameters *pParameters,
    NvOdmImagerSensorMode *pSelectedMode,
    SetModeParameters *pResult)
{
    NvBool Status = NV_FALSE;
    SensorBayerContext *pContext =
        (SensorBayerContext*)hImager->pSensor->pPrivateContext;
    NvU32 Index;
    ModeDependentSettings *pModeSettings;

    DEBUGMSG(ZONE_FUNC, (_T("+SensorYuv_SetMode\n")));

    Status = SensorBayerChooseModeIndex(pContext, pParameters->Resolution,
                                        &Index);

    // No match found
    if (!Status)
        return NV_FALSE;

    if (pSelectedMode)
    {
        *pSelectedMode = g_SensorBayerSetModeSequenceList[Index].Mode;
    }

    if (pContext->ModeIndex == Index)
        return NV_TRUE;

    // TODO: Do we need to set a lower coarse time for new framelenghtline?
    // I don't think so because the set mode sequence reset the sensor.

    // i2c writes for the set mode sequence
    //Status = WriteI2CSequence(&pContext->I2c,
    //             g_SensorBayerSetModeSequenceList[Index].pSequence,
    //             pContext->SensorInitialized && pContext->FastSetMode);
	//WriteToI2C(g_SensorBayerSetModeSequenceList[Index].pSequence);
    SensorI2C_WriteSequence(g_SensorBayerSetModeSequenceList[Index].pSequence);

    if (!Status)
        return NV_FALSE;
    /**
     * the following is Phase 2.
     */
    pContext->SensorInitialized = NV_TRUE;

    // Update sensor context after set mode
    //NV_ASSERT(pContext->SensorInputClockkHz > 0);

    // These hardcoded numbers are from the set mode sequence and this formula
    // is based on this sensor. Sensor Dependent.
    pContext->PllMult = SENSOR_BAYER_DEFAULT_PLL_MULT;
    pContext->PllPreDiv = SENSOR_BAYER_DEFAULT_PLL_PRE_DIV;
    pContext->PllVtPixDiv = SENSOR_BAYER_DEFAULT_PLL_VT_PIX_DIV;
    pContext->PllVtSysDiv = SENSOR_BAYER_DEFAULT_PLL_VT_SYS_DIV;
    pContext->VtPixClkFreqHz =
        (NvU32)((pContext->SensorInputClockkHz * 1000 * pContext->PllMult) /
                (pContext->PllPreDiv * pContext->PllVtPixDiv *
                 pContext->PllVtSysDiv));

    pModeSettings =
        (ModeDependentSettings*)g_SensorBayerSetModeSequenceList[Index].
                                pModeDependentSettings;

    pContext->LineLength = pModeSettings->LineLength;
    pContext->FrameLength = pModeSettings->FrameLength;
    pContext->CoarseTime = pModeSettings->CoarseTime;

    pContext->Exposure = (NvF32)(pContext->CoarseTime * pContext->LineLength +
                                 SENSOR_BAYER_DEFAULT_EXPOSURE_FINE) /
                         (NvF32)pContext->VtPixClkFreqHz;
    pContext->MaxExposure = (NvF32)((pContext->FrameLength-1) * pContext->LineLength +
                                 SENSOR_BAYER_DEFAULT_EXPOSURE_FINE) /
                         (NvF32)pContext->VtPixClkFreqHz;
    pContext->MinExposure = (NvF32)(1 * pContext->LineLength +
                                 SENSOR_BAYER_DEFAULT_EXPOSURE_FINE) /
                         (NvF32)pContext->VtPixClkFreqHz;

    pContext->MaxFrameLength = SENSOR_BAYER_DEFAULT_MAX_FRAME_LENGTH;
    pContext->MinFrameLength = pModeSettings->MinFrameLength;

	pContext->FrameRate = g_SensorBayerSetModeSequenceList[Index].Mode.PeakFrameRate;
	pContext->MinFrameRate = 0;
	pContext->MaxFrameRate = 120;
	
	//DAN
    pContext->FrameRate = (NvF32)pContext->VtPixClkFreqHz /
                          (NvF32)(pContext->FrameLength *
                                  pContext->LineLength); 
    //pContext->MaxFrameRate = pContext->FrameRate;
    //pContext->MinFrameRate = (NvF32)pContext->VtPixClkFreqHz /
    //                         (NvF32)(SENSOR_BAYER_DEFAULT_MAX_FRAME_LENGTH *
    //                                 pContext->LineLength);

    pContext->Gains[0] = SENSOR_BAYER_DEFAULT_GAIN_R;
    pContext->Gains[1] = SENSOR_BAYER_DEFAULT_GAIN_GR;
    pContext->Gains[2] = SENSOR_BAYER_DEFAULT_GAIN_GB;
    pContext->Gains[3] = SENSOR_BAYER_DEFAULT_GAIN_B;



    SensorBayer_WriteExposure(pContext, &pParameters->Exposure);
    SensorBayer_WriteGains(pContext, pParameters->Gains);

    // Do we need to slow down the frame rate?
    //if (pContext->RequestedMaxFrameRate > 0.0f)
    //{
    //    NvU32 RequestedMinFrameLengthLines = 
    //        (NvU32)(pContext->VtPixClkFreqHz / (pContext->LineLength *
    //                pContext->RequestedMaxFrameRate));

    //    if (RequestedMinFrameLengthLines > pContext->MaxFrameLength)
    //        RequestedMinFrameLengthLines = pContext->MaxFrameLength;

    //    if (pContext->FrameLength < RequestedMinFrameLengthLines)
    //    {
    //        NVODM_WRITE_RETURN_ON_ERROR(&pContext->I2c,
    //                0x0340, (RequestedMinFrameLengthLines >> 8) & 0xFF);
    //        NVODM_WRITE_RETURN_ON_ERROR(&pContext->I2c,
    //                0x0341, RequestedMinFrameLengthLines & 0xFF);
    //        
    //        pContext->FrameLength = RequestedMinFrameLengthLines;
    //    }
    //}


    if (pResult)
    {
        pResult->Resolution = g_SensorBayerSetModeSequenceList[Index].
                              Mode.ActiveDimensions;
        pResult->Exposure = pContext->Exposure;
        memcpy(pResult->Gains, &pContext->Gains, sizeof(NvF32) * 4);
    }

    /**
     * Phase 2 ends here.
     */

    pContext->ModeIndex = Index;

    //Wait 2 frames for gains/exposure to take effect.
    Sleep((NvU32)(2000.0 / (NvF32)(pContext->FrameRate)));
    
    DEBUGMSG(ZONE_FUNC, (_T("-SensorYuv_SetMode\n")));

    return Status;
}


/**
 * SensorBayer_SetPowerLevel. Phase 1
 * Set the sensor's power level
 */
static NvBool
SensorBayer_SetPowerLevel(
    NvOdmImagerHandle hImager,
    NvOdmImagerPowerLevel PowerLevel)
{
    NvBool Status = NV_TRUE;

    SensorBayerContext *pContext =
        (SensorBayerContext*)hImager->pSensor->pPrivateContext;
    //const NvOdmPeripheralConnectivity *pConnections;
    
    DEBUGMSG(ZONE_FUNC, (_T("+SensorYuv_SetPowerLevel\n")));

    if (pContext->PowerLevel == PowerLevel)
        return NV_TRUE;

    ////NV_ASSERT(hImager->pSensor->pConnections);
    //pConnections = hImager->pSensor->pConnections;

    switch(PowerLevel)
    {
        case NvOdmImagerPowerLevel_On:
 /*           Status = SetPowerLevelWithPeripheralConnectivity(pConnections,
                         &pContext->I2c, &pContext->GpioConfig,
                         NvOdmImagerPowerLevel_On);
            if (!Status) {
                return NV_FALSE;
            }*/

            break;

        case NvOdmImagerPowerLevel_Standby:

            //Status = SetPowerLevelWithPeripheralConnectivity(pConnections,
            //             &pContext->I2c, &pContext->GpioConfig,
            //             NvOdmImagerPowerLevel_Standby);

            //break;
        case NvOdmImagerPowerLevel_Off:

            // // Set the sensor to standby mode
            //NVODM_WRITE_RETURN_ON_ERROR(&pContext->I2c, 0x0100, 0);

            //Status = SetPowerLevelWithPeripheralConnectivity(pConnections,
            //             &pContext->I2c, &pContext->GpioConfig,
            //             NvOdmImagerPowerLevel_Off);
            //
            //// wait for the standby 0.1 sec should be enough
            //NvOdmOsWaitUS(100000); 

            break;
        default:
            //NV_ASSERT("!Unknown power level\n");
            Status = NV_FALSE;
            break;
    }

    if (Status)
        pContext->PowerLevel = PowerLevel;

    DEBUGMSG(ZONE_FUNC, (_T("-SensorYuv_SetPowerLevel\n")));

    return Status;
}


static TCHAR* Float2String(NvF32 f)
{
    static TCHAR tmpStr[32];
    wsprintf(tmpStr, L"%f", f);
    return tmpStr;
}

/**
 * SensorBayer_SetParameter. Phase 2.
 * Set sensor specific parameters. 
 * For Phase 1. This can return NV_TRUE.
 */
static NvBool
SensorBayer_SetParameter(
    NvOdmImagerHandle hImager,
    NvOdmImagerParameter Param,
    NvS32 SizeOfValue,
    const void *pValue)
{
    NvBool Status = NV_TRUE;
    SensorBayerContext *pContext =
        (SensorBayerContext*)hImager->pSensor->pPrivateContext;
    BOOL bDebugEnable = TRUE;

#if DEBUG //NvOdmImagerParameter_SensorFrameRate is called too often, output it only once
    {
        static NvOdmImagerParameter OldParam = 0;

        if(Param==NvOdmImagerParameter_SensorTriggerStill)
        {
            if(OldParam&1)
                bDebugEnable=FALSE;
            OldParam |= 1;
        }
        /*if(Param==xxd)
        {
            if(OldParam&2)
                bDebugEnable=FALSE;
            OldParam |= 2;
        }*/
    }
#endif

    DEBUGMSG(ZONE_FUNC && bDebugEnable, (_T("+SensorYuv_SetParameter(%S)\n"), GetParamName(Param)));

    switch(Param)
    {
        case NvOdmImagerParameter_SensorTriggerStill:
        {
            static NvU32 t0=-1;
            NvU32 t1;
            static NvU32 frames=0;

            if(t0==-1)
                t0=GetTickCount();

            frames++;
            t1 = GetTickCount();

            if(t1-t0>=1000)
            {
                DEBUGMSG(ZONE_INFO, (_T("Average fps: %d\n"), frames));
                t0 = t1;
                frames=0;
            }
            
            break;
        }
        // Phase 2.
        case NvOdmImagerParameter_SensorExposure:
			if(SizeOfValue!=sizeof(NvF32)){
				DEBUGMSG(ZONE_ERR, (L"ERROR: SensorBayer_SetParameter (SensorExposure): SizeOfValue does not match \n\r"));
				return FALSE;
			}

            Status = SensorBayer_WriteExposure(pContext, (NvF32*)pValue);
            DEBUGMSG(ZONE_INFO, (_T("Exposure: %s us\n"), Float2String(*(NvF32*)pValue) ));
            break;

        // Phase 2.
        case NvOdmImagerParameter_SensorGain:
			if(SizeOfValue!=4*sizeof(NvF32)){
				DEBUGMSG(ZONE_ERR, (L"ERROR: SensorBayer_SetParameter (SensorGain): SizeOfValue does not match \n\r"));
				return FALSE;
			}

            Status = SensorBayer_WriteGains(pContext, pValue);
            DEBUGMSG(ZONE_INFO, (_T("Gain: %s\n"), Float2String(*(NvF32*)pValue) ));
            break;

        // Phase 2.
        case NvOdmImagerParameter_SensorInputClock:
			if(SizeOfValue!=sizeof(NvOdmImagerClockProfile)){
				DEBUGMSG(ZONE_ERR, (L"ERROR: SensorBayer_SetParameter (SensorInputClock): SizeOfValue does not match \n\r"));
				return FALSE;
			}
            
            Status = SensorBayer_SetInputClock(
                      pContext,
                      ((NvOdmImagerClockProfile*)pValue));
            break;

        // Phase 3.
        case NvOdmImagerParameter_OptimizeResolutionChange:
			//if(SizeOfValue!=sizeof(BOOL)){
			//	NKDbgPrintfW(L"ERROR: SensorBayer_SetParameter (OptimizeResolutionChange): SizeOfValue does not match \n\r");
			//	return FALSE;
			//}

            pContext->FastSetMode = *((NvBool*)pValue);
            break;

        // This is optional but nice to have.
        //case NvOdmImagerParameter_SelfTest:
        //    //TODO: Not Implemented.
        //    break;
        // Phase 3.
        case NvOdmImagerParameter_Reset:
            Status = NV_FALSE;
            break;

        case NvOdmImagerParameter_MaxSensorFrameRate:
			if(SizeOfValue!=sizeof(NvF32)){
				DEBUGMSG(ZONE_ERR, (L"ERROR: SensorBayer_SetParameter (MaxSensorFrameRate): SizeOfValue does not match \n\r"));
				Status = NV_FALSE;
			}

            pContext->RequestedMaxFrameRate = *((NvF32*)pValue);
            break;

        default:
            DEBUGMSG(ZONE_FUNC && ZONE_WARN && bDebugEnable, (L"Not Implemented\n"));
            Status = NV_FALSE;
            break;
    }
    
    DEBUGMSG(ZONE_FUNC && bDebugEnable, (_T("-SensorYuv_SetParameter(%d)\n"), Status));

    return Status;
}

/**
 * SensorBayer_GetParameter. Phase 1.
 * Get sensor specific parameters
 */
static NvBool
SensorBayer_GetParameter(
    NvOdmImagerHandle hImager,
    NvOdmImagerParameter Param,
    NvS32 SizeOfValue,
    void *pValue)
{
	 NvBool Status = NV_TRUE; 
    NvF32 *pFloatValue = pValue;
    SensorBayerContext *pContext =
        (SensorBayerContext*)hImager->pSensor->pPrivateContext;
    NvU32 *pUIntValue = pValue;
    BOOL bDebugEnable = TRUE;

#if DEBUG //NvOdmImagerParameter_SensorFrameRate is called too often, output it only once
    {
        static NvOdmImagerParameter OldParam = 0;

        if(Param==NvOdmImagerParameter_SensorFrameRate)
        {
            if(OldParam&1)
                bDebugEnable=FALSE;
            OldParam |= 1;
        }
       /* if(Param==NvOdmImagerParameter_SensorExposureLimits)
        {
            if(OldParam&2)
                bDebugEnable=FALSE;
            OldParam |= 2;
        }*/
    }
#endif

    DEBUGMSG(ZONE_FUNC && bDebugEnable, (L"+SensorYuv_GetParameter(%S)\n", GetParamName(Param)));

    switch(Param)
    {
        // Phase 1.
        case NvOdmImagerParameter_CalibrationData:
 /*           CHECK_PARAM_SIZE_RETURN_MISMATCH(SizeOfValue,
                sizeof(NvOdmImagerCalibrationData));*/
            {
                NvOdmImagerCalibrationData *pCalibration =
                        (NvOdmImagerCalibrationData*)pValue;
                pCalibration->CalibrationData = pSensorCalibrationData;
                pCalibration->NeedsFreeing = NV_FALSE;
            }
            break;

        // Phase 1, it can return min = max = 1.0f
        // Phase 2, return the real numbers.
        case NvOdmImagerParameter_SensorExposureLimits:
            //CHECK_PARAM_SIZE_RETURN_MISMATCH(SizeOfValue, 2 * sizeof(NvF32));

            pFloatValue[0] = pContext->MinExposure;
            pFloatValue[1] = pContext->MaxExposure;
            break;

        // Phase 1, it can return min = max = 1.0f
        // Phase 2, return the real numbers.
        case NvOdmImagerParameter_SensorGainLimits:
            //CHECK_PARAM_SIZE_RETURN_MISMATCH(SizeOfValue, 2 * sizeof(NvF32));
            pFloatValue[0] = pContext->MinGain;
            pFloatValue[1] = pContext->MaxGain;

            break;

        // Get the sensor status. This is optional but nice to have.
#if 0
        case NvOdmImagerParameter_DeviceStatus:
            //CHECK_PARAM_SIZE_RETURN_MISMATCH(SizeOfValue,
            //    sizeof(NvOdmImagerDeviceStatus));
            {
                NvOdmImagerDeviceStatus *pStatus;
                pStatus = (NvOdmImagerDeviceStatus *)pValue;
                {
                    //NvOdmImagerI2cConnection *pI2c = &pContext->I2c;
                    // Pick your own useful registers to use for debug
                    // If the camera hangs, these register values are printed
                    // Sensor Dependent.
/*
                    NvOdmImagerI2cRead(pI2c, 0x0000, &pStatus->Values[0]);
                    NvOdmImagerI2cRead(pI2c, 0x303B, &pStatus->Values[1]);
                    NvOdmImagerI2cRead(pI2c, 0x303C, &pStatus->Values[2]);
                    NvOdmImagerI2cRead(pI2c, 0x306A, &pStatus->Values[3]);
*/
                    pStatus->Count = 4;
                }
            }
            break;
#endif
        case NvOdmImagerParameter_SensorFrameRate:
            //CHECK_PARAM_SIZE_RETURN_MISMATCH(SizeOfValue, sizeof(NvF32));
            {
                pFloatValue[0] = pContext->FrameRate;
            }
            Status = NV_TRUE;
            break;

        case NvOdmImagerParameter_SensorFrameRateLimits:
            //CHECK_PARAM_SIZE_RETURN_MISMATCH(SizeOfValue,
            //    2 * sizeof(NvF32));
            pFloatValue[0] = pContext->MinFrameRate;
            pFloatValue[1] = pContext->MaxFrameRate;
            break;

        case NvOdmImagerParameter_RegionUsedByCurrentResolution:
            if (SizeOfValue != sizeof(NvOdmImagerRegion)) {
                //NV_ASSERT(!"Bad size");
                return NV_FALSE;
            }
            if (pContext->ModeIndex == (pContext->NumModes+1)) { // Invalid mode
                // No info available until we've actually set a mode.
                return NV_FALSE;
            }
            {
                NvOdmImagerRegion *pRegion = (NvOdmImagerRegion *)pValue;
                pRegion->RegionStart.x = 0;
                pRegion->RegionStart.y = 0;
                if (pContext->ModeIndex == 0) { // Quarter res
                    pRegion->xScale = 2;
                    pRegion->yScale = 2;
                } else { // Should be Full res.
                    pRegion->xScale = 1;
                    pRegion->yScale = 1;
                }
                return NV_TRUE;
            }
            break;
#if 0
        case NvOdmImagerParameter_SensorFrameRateLimitsAtResolution:
            //CHECK_PARAM_SIZE_RETURN_MISMATCH(SizeOfValue,
            //    sizeof(NvOdmImagerFrameRateLimitAtResolution));
            {
                NvOdmImagerFrameRateLimitAtResolution *pData;
                NvU32 Index;
                NvBool MatchFound = NV_FALSE;
                ModeDependentSettings *pModeSettings = NULL;

                pData = (NvOdmImagerFrameRateLimitAtResolution *)pValue;
                MatchFound =
                    SensorBayerChooseModeIndex(pContext, pData->Resolution,
                    &Index);

                if (!MatchFound)
                {
                    pData->MinFrameRate = 0.0f;
                    pData->MaxFrameRate = 0.0f;
                    break;
                }

                pModeSettings = (ModeDependentSettings*)
                    g_SensorBayerSetModeSequenceList[Index].
                    pModeDependentSettings;

				pData->MaxFrameRate = 30.0;
				pData->MinFrameRate = 0.0;
				//DAN
                //pData->MaxFrameRate = (NvF32)pContext->VtPixClkFreqHz /
                //                      (NvF32)(pModeSettings->FrameLength *
                //                              pModeSettings->LineLength);
                //pData->MinFrameRate =
                //    (NvF32)pContext->VtPixClkFreqHz /
                //    (NvF32)(SENSOR_BAYER_DEFAULT_MAX_FRAME_LENGTH *
                //            pModeSettings->LineLength);

            }
            break;
#endif

        // Phase 3.
        // Get the override config data.
        case NvOdmImagerParameter_CalibrationOverrides:
			//DAN
			NKDbgPrintfW(L"DAN TO DO NvOdmImagerParameter_CalibrationOverrides\n\r");
            Status = NV_FALSE;
			/*
            CHECK_PARAM_SIZE_RETURN_MISMATCH(SizeOfValue,
                sizeof(NvOdmImagerCalibrationData));
            {
                NvOdmImagerCalibrationData *pCalibration =
                    (NvOdmImagerCalibrationData*)pValue;

                pCalibration->CalibrationData =
                    LoadOverridesFile(pOverrideFiles,
                        (sizeof(pOverrideFiles)/sizeof(pOverrideFiles[0])));
                pCalibration->NeedsFreeing = (pCalibration->CalibrationData != NULL);
                Status = pCalibration->NeedsFreeing;
            }
			*/
            break;

#if 0
        case NvOdmImagerParameter_SensorExposureLatchTime:
            //CHECK_PARAM_SIZE_RETURN_MISMATCH(SizeOfValue,
            //    sizeof(NvU32));
                *pUIntValue = 1;
            break;
#endif
        // Phase 3.
        default:
            //NV_ASSERT("!Not Implemented");
            Status = NV_FALSE;
    }

    DEBUGMSG(ZONE_FUNC && bDebugEnable, (L"-SensorYuv_GetParameter(%d)\n", Status));

    return Status;
}

/**
 * SensorBayer_GetPowerLevel. Phase 3.
 * Get the sensor's current power level
 */
static void
SensorBayer_GetPowerLevel(
    NvOdmImagerHandle hImager,
    NvOdmImagerPowerLevel *pPowerLevel)
{
    SensorBayerContext *pContext =
        (SensorBayerContext*)hImager->pSensor->pPrivateContext;

    DEBUGMSG(ZONE_FUNC, (L"+SensorYuv_GetPowerLevel\n"));

    *pPowerLevel = pContext->PowerLevel;

    DEBUGMSG(ZONE_FUNC, (L"-SensorYuv_GetPowerLevel\n"));
}

/**
 * SensorBayer_GetHal. Phase 1.
 * return the hal functions associated with sensor bayer
 */
NvBool SensorPlugin_GetHal_Bayer_MT9P031(NvOdmImagerHandle hImager)
{
    if (!hImager || !hImager->pSensor)
        return NV_FALSE;

    hImager->pSensor->pfnOpen = SensorBayer_Open;
    hImager->pSensor->pfnClose = SensorBayer_Close;
    hImager->pSensor->pfnGetCapabilities = SensorBayer_GetCapabilities;
    hImager->pSensor->pfnListModes = SensorBayer_ListModes;
    hImager->pSensor->pfnSetMode = SensorBayer_SetMode;
    hImager->pSensor->pfnSetPowerLevel = SensorBayer_SetPowerLevel;
    hImager->pSensor->pfnGetPowerLevel = SensorBayer_GetPowerLevel;
    hImager->pSensor->pfnSetParameter = SensorBayer_SetParameter;
    hImager->pSensor->pfnGetParameter = SensorBayer_GetParameter;

    return NV_TRUE;
}
