/*
 * Copyright (c) 2008 NVIDIA Corporation.  All rights reserved.
 *
 * NVIDIA Corporation and its licensors retain all intellectual property
 * and proprietary rights in and to this software, related documentation
 * and any modifications thereto.  Any use, reproduction, disclosure or
 * distribution of this software and related documentation without an express
 * license agreement from NVIDIA Corporation is strictly prohibited.
 */

//#include "sensor_bayer_MT9P031.h"
//#include "focuser.h"
//#include "imager_util.h"
//#include "nvodm_query_gpio.h"
#include "nvodm_imager.h"
#include "nvodm_imager_int.h"
#include "imager_hal.h" //shard ressources with driver
#include "I2clib.h"
#include "GPIOLib.h"

#include "sensor_bayer_camera_config_min.h"

/****************************************************************************
 * Implementing for a New Sensor:
 * 
 * When implementing for a new sensor from this template, there are four 
 * phases:
 * Phase 0: Gather sensor information. In phase 0, we should gather the sensor
 *          information for bringup such as sensor capabilities, set mode i2c
 *          write sequence, NvOdmIoAddress in nvodm_query_discovery_imager.h,
 *          and so on.
 * Phase 1: Bring up a sensor. After phase 1, we should be able to do a basic
 *          still capture with minimal interactions between camera driver
 *          and the sensor
 * Phase 2: Calibrate a sensor. After phase 2, we should be able to calibrate
 *          a sensor. Calibration will need some interactions between camera
 *          driver and the sensor such as setting exposure, gains, and so on.
 * Phase 3: Fully functioning sensor. After phase 3, the sensor should be fully
 *          functioning as described in nvodm_imager.h
 *
 * To help implementing for a new sensor, the template code will be marked as
 * Phase 0, 1, 2, or 3. 
 ****************************************************************************/

//For I2C
#define SEQUENCE_WAIT_MS                (0xFFFE)
#define SEQUENCE_END                    (0xFFFF)
#define SEQUENCE_FAST_SETMODE_START     (0xFFFD)
#define SEQUENCE_FAST_SETMODE_END       (0xFFFC)


/**
 * Sensor default settings. Phase 2. Sensor Dependent.
 * #defines make the code more readable and helps maintainability
 * The template tries to keep the list reduced to just things that
 * are used more than once.  Ideally, all numbers in the code would use
 * a #define.
 */
#define SENSOR_BAYER_DEFAULT_EXPOSURE_FINE      (0x0190)
#define SENSOR_BAYER_DEFAULT_PLL_MULT           (0x0001) //DAN
#define SENSOR_BAYER_DEFAULT_PLL_PRE_DIV        (0x0001) //DAN
#define SENSOR_BAYER_DEFAULT_PLL_VT_PIX_DIV     (0x0001) //DAN
#define SENSOR_BAYER_DEFAULT_PLL_VT_SYS_DIV     (0x0001)
#define SENSOR_BAYER_DEFAULT_MAX_GAIN           (63)
#define SENSOR_BAYER_DEFAULT_MIN_GAIN           (8)
#define SENSOR_BAYER_DEFAULT_GAIN_R             (8)
#define SENSOR_BAYER_DEFAULT_GAIN_GR            (8)
#define SENSOR_BAYER_DEFAULT_GAIN_GB            (8)
#define SENSOR_BAYER_DEFAULT_GAIN_B             (8)
#define SENSOR_BAYER_DEFAULT_MAX_FRAME_LENGTH   (0xFFFF)
#define SENSOR_BAYER_DEFAULT_MIN_FRAME_LENGTH   (1976+21)

#define SENSOR_BAYER_DEFAULT_LINE_LENGTH_FULL       (0x1220)
#define SENSOR_BAYER_DEFAULT_FRAME_LENGTH_FULL      (0x07D0)
#define SENSOR_BAYER_DEFAULT_EXPOSURE_COARSE_FULL   (0x07D0)

#define SENSOR_BAYER_DEFAULT_LINE_LENGTH_QUAR       (0x0910)
#define SENSOR_BAYER_DEFAULT_FRAME_LENGTH_QUAR      (0x03E8)
#define SENSOR_BAYER_DEFAULT_EXPOSURE_COARSE_QUAR   (0x03E8)

#define SENSOR_GAIN_TO_F32(x)     (128.f / (128.f - (NvF32)(x)))
#define SENSOR_F32_TO_GAIN(x)     ((NvU32)(128.f - (128.f / (x))))


//#define SODIMM_PIN_SENSOR_ENABLE	(77)	//Low active 
#define SODIMM_PIN_SENSOR_RESET		(69)	//Low active
#define I2C_ADDRESS					(0x30)
#define SWAP_BYTES(w)	( (w>>8)|((w&0xff)<<8) )


/**
 * ----------------------------------------------------------
 *  Start of Phase 0 code
 * ----------------------------------------------------------
 */

/**
 * Sensor Specific Variables/Defines. Phase 0.
 *
 * If a new sensor code is created from this template, the following
 * variable/defines should be created for the new sensor specifically.
 * 1. A sensor capabilities (NvOdmImagerCapabilities) for this specific sensor
 * 2. A sensor set mode sequence list  (an array of SensorSetModeSequence)
 *    consisting of pairs of a sensor mode and corresponding set mode sequence
 * 3. A set mode sequence for each mode defined in 2.
 */

/**
 * Sensor bayer's capabilities. Phase 0. Sensor Dependent.
 * This is the first thing the camera driver requests. The information here
 * is used to setup the proper interface code (CSI, VIP, pixel type), start
 * up the clocks at the proper speeds, and so forth.
 * For Phase 0, one could start with these below as an example for a bayer
 * sensor and just update clocks from what is given in the reference documents
 * for that device. The FAE may have specified the clocks when giving the
 * reccommended i2c settings.
 */
static NvOdmImagerCapabilities g_SensorBayerCaps =
{
    "MT9P031",  // string identifier, used for debugging
    NvOdmImagerSensorInterface_Parallel8, // interconnect type
    {
    //NvOdmImagerPixelType_BayerRGGB,
    //NvOdmImagerPixelType_BayerBGGR,
    //NvOdmImagerPixelType_BayerGRBG,
    NvOdmImagerPixelType_BayerGBRG,
    //NvOdmImagerPixelType_UYVY,
	},
    NvOdmImagerOrientation_0_Degrees, // rotation with respect to main display
    NvOdmImagerDirection_Away,          // toward or away the main display
    48000, // Initial sensor clock (in kHz).
    { {48000, 1.0}, {0,0} }, // preferred clock profile
    {
        NvOdmImagerSyncEdge_Rising, // hsync polarity
        NvOdmImagerSyncEdge_Rising, // vsync polarity
        NV_TRUE,                   // uses mclk on vgp0
    }, // parallel settings (VIP)
    { 0 }, // serial settings (CSI) (not used)
    { 16, 7 }, // min blank time, shouldn't need adjustment
    0, // preferred mode, which resolution to use on startup
    0, // FOCUSER_GUID, only set if focuser code available
    0, // FLASH_GUID, only set if flash device code is available
    NVODM_IMAGER_CAPABILITIES_END
};

 

static DWORD WriteToI2C(DevCtrlReg16* SetModeSequence)
{
	char data[10];
	
	DevCtrlReg16 *ptr = SetModeSequence;
	DWORD ret= TRUE;

	if (I2CInit())
	{
		if(GetI2CLock(500)){
			//I2CSetSpeed(I2C_100KBPS);				// Currently, this sets LastError=1210 on Tegra
			
			while(ptr->RegAddr!= SEQUENCE_END)
			{
				if(ptr->RegAddr == SEQUENCE_WAIT_MS){
					Sleep(1000*ptr->RegValue);
				}
				else{
					//switch bytes
					data[1] = (char)ptr->RegValue;
					data[0] = (char)(ptr->RegValue>>8);
					if(!I2CBurstWrite(I2C_ADDRESS, data, (char)ptr->RegAddr, 2))
					{
						NKDbgPrintfW(L"WriteToI2C I2CBurstWrite faild\n\r");
						ret= FALSE;
						break;
					}
				}
				ptr += 1;
			}
			ReleaseI2CLock();
		}
		else{
			ret= FALSE;
			NKDbgPrintfW(L"WriteToI2C GetI2CLock faild\n\r");
		}
		I2CDeInit();
	}
	else{
		ret= FALSE;
		NKDbgPrintfW(L"WriteToI2C I2CInit faild\n\r");
	}
	return ret;

}

//;-------------------------------------------------------------------
void SetOmnivisionReg(BYTE reg, BYTE val)
{
	
	I2CBurstWrite(I2C_ADDRESS, (char*)&val, (short)reg, sizeof(BYTE)); // this does not work if REG is 0xff
//		I2CWrite(I2C_ADDRESS, &reg, 1, FALSE);	// send out address,register, NO stop
//		I2CWrite(255, &val, 1, TRUE);					// send byte to be written, STOP
}

//;-------------------------------------------------------------------
BYTE GetOmnivisionReg(BYTE reg)
{
	BYTE val;

	I2CBurstWrite(I2C_ADDRESS, (char*)&reg, I2C_OFFSET_NONE, sizeof(BYTE));
	I2CBurstRead(I2C_ADDRESS, (char*)&val, I2C_OFFSET_NONE, sizeof(val));
	//I2CBurstRead(I2C_ADDRESS, (char*)&val, reg, sizeof(BYTE)); // this does not work if REG is 0xff
	return val;
}




typedef struct ModeDependentSettingsRec
{
    NvU32 LineLength;
    NvU32 FrameLength;
    NvU32 MinFrameLength;
    NvU32 CoarseTime;
} ModeDependentSettings;

static ModeDependentSettings ModeDependentSettings_1316x988 = 
{
    SENSOR_BAYER_DEFAULT_LINE_LENGTH_QUAR,
    SENSOR_BAYER_DEFAULT_FRAME_LENGTH_QUAR,
    SENSOR_BAYER_DEFAULT_FRAME_LENGTH_QUAR,
    SENSOR_BAYER_DEFAULT_EXPOSURE_COARSE_QUAR,
};

static ModeDependentSettings ModeDependentSettings_2632x1976 = 
{
    SENSOR_BAYER_DEFAULT_LINE_LENGTH_FULL,
    SENSOR_BAYER_DEFAULT_FRAME_LENGTH_FULL,
    SENSOR_BAYER_DEFAULT_FRAME_LENGTH_FULL,
    SENSOR_BAYER_DEFAULT_EXPOSURE_COARSE_FULL,
};

static ModeDependentSettings ModeDependentSettings_640x480 = 
{
    SENSOR_BAYER_DEFAULT_LINE_LENGTH_FULL,
    SENSOR_BAYER_DEFAULT_FRAME_LENGTH_FULL,
    SENSOR_BAYER_DEFAULT_FRAME_LENGTH_FULL,
    SENSOR_BAYER_DEFAULT_EXPOSURE_COARSE_FULL,
};

static DevCtrlReg16 SetModeSequence_640x480[] =
{
		{SEQUENCE_END,0}
};

/**
 * A list of sensor mode and its SetModeSequence. Phase 0. Sensor Dependent.
 * When SetMode is called, it searches the list for desired resolution and
 * frame rate and write the SetModeSequence through i2c.
 * For Phase 3, more resolutions will be added to this table, allowing
 * for peak performance during Preview, D1 and 720P encode.
 */
static SensorSetModeSequence g_SensorBayerSetModeSequenceList[] =
{
     // WxH         x,y    fps    set mode sequence
   {{{640, 480}, {0, 0}, 30,}, SetModeSequence_640x480,&ModeDependentSettings_640x480},
//	{{{640, 480}, {2, 2}, 30,}, SetModeSequence_640x480_TestMode,&ModeDependentSettings_640x480},
//    {{{2632, 1976}, {2, 2}, 1.5,}, SetModeSequence_2632x1976,&ModeDependentSettings_2632x1976},
//	{{{2592, 1944}, {1, 1}, 30,}, SetModeSequence_2592x1944,&ModeDependentSettings_2632x1976}
};


/**
 * ----------------------------------------------------------
 *  Start of Phase 1, Phase 2, and Phase 3 code
 * ----------------------------------------------------------
 */

/**
 * Sensor bayer's private context. Phase 1.
 */
typedef struct SensorBayerContextRec
{
    // Phase 1 variables.
    //NvOdmImagerI2cConnection I2c;
    NvU32 ModeIndex;
    NvU32 NumModes;
    NvOdmImagerPowerLevel PowerLevel;


    // Phase 2 variables.
    NvBool SensorInitialized;

    NvU32 SensorInputClockkHz; // mclk (extclk)
    NvF32 SensorClockRatio; // pclk/mclk

    NvF32 Exposure;
    NvF32 MaxExposure;
    NvF32 MinExposure;

    NvF32 Gains[4];
    NvF32 MaxGain;
    NvF32 MinGain;

    NvF32 FrameRate;
    NvF32 RequestedMaxFrameRate; // requested limit
    NvF32 MinFrameRate; // sensor's limit
    NvF32 MaxFrameRate; // sensor's limit

    // Phase 2 variables. Sensor Dependent.
    // These are used to set or get exposure, frame rate, and so on.
	
    NvU32 PllMult;
    NvU32 PllPreDiv;
    NvU32 PllVtPixDiv;
    NvU32 PllVtSysDiv;

    NvU32 CoarseTime;
    NvU32 VtPixClkFreqHz;

    NvU32 LineLength;
    NvU32 FrameLength;

    NvU32 MaxFrameLength;
    NvU32 MinFrameLength;

    // Phase 3.
    NvBool FastSetMode;


}SensorBayerContext;


/**
 * Static functions
 * Some compilers require these to be declared upfront.
 */
static NvBool SensorBayer_Open(NvOdmImagerHandle hImager);

static void SensorBayer_Close(NvOdmImagerHandle hImager);

static void
SensorBayer_GetCapabilities(
    NvOdmImagerHandle hImager,
    NvOdmImagerCapabilities *pCapabilities);

static void
SensorBayer_ListModes(
    NvOdmImagerHandle hImager,
    NvOdmImagerSensorMode *pModes,
    NvS32 *pNumberOfModes);

static NvBool
SensorBayer_SetMode(
    NvOdmImagerHandle hImager,
    const SetModeParameters *pParameters,
    NvOdmImagerSensorMode *pSelectedMode,
    SetModeParameters *pResult);

static NvBool
SensorBayer_SetPowerLevel(
    NvOdmImagerHandle hImager,
    NvOdmImagerPowerLevel PowerLevel);

static NvBool
SensorBayer_SetParameter(
    NvOdmImagerHandle hImager,
    NvOdmImagerParameter Param,
    NvS32 SizeOfValue,
    const void *pValue);

static NvBool
SensorBayer_GetParameter(
    NvOdmImagerHandle hImager,
    NvOdmImagerParameter Param,
    NvS32 SizeOfValue,
    void *pValue);

static void
SensorBayer_GetPowerLevel(
    NvOdmImagerHandle hImager,
    NvOdmImagerPowerLevel *pPowerLevel);

static NvBool
SensorBayer_WriteFramerate(
        SensorBayerContext *pContext,
        NvF32 *pFrameRate,
        NvBool WriteNewFrameRate);

/**
 * SensorBayer_WriteExposure. Phase 2. Sensor Dependent.
 * Calculate and write the values of the sensor registers for the new
 * exposure time.
 * Without this, calibration will not be able to capture images
 * at various brightness levels, and the final product won't be able
 * to compensate for bright or dark scenes. 
 */
static NvBool
SensorBayer_WriteExposure(
    SensorBayerContext *pContext,
    const NvF32 *pNewExposureTime)
{

	NKDbgPrintfW(L"SensorBayer_WriteExposure NOT IMPLEMENTED\n\r");
    //NvF32 Freq = (NvF32)pContext->VtPixClkFreqHz;
    //NvF32 LineLength = (NvF32)pContext->LineLength;
    //NvF32 ExpTime = *pNewExposureTime;
    //NvU32 NewCoarseTime = 0;
    //NvU32 NewFrameLength = 0;


    //if (*pNewExposureTime > pContext->MaxExposure ||
    //    *pNewExposureTime < pContext->MinExposure)
    //    return NV_FALSE;

    //// Here, we have to decide if the new exposure time is valid
    //// based on the sensor and current sensor settings.
    //NewCoarseTime = (NvU32)((Freq * ExpTime) / LineLength);
    //NewFrameLength = NewCoarseTime + 1;

    //// Consider requested max frame rate,
    //if (pContext->RequestedMaxFrameRate > 0.0f)
    //{
    //    NvU32 RequestedMinFrameLengthLines = 0;
    //    RequestedMinFrameLengthLines =
    //       (NvU32)(Freq / (LineLength * pContext->RequestedMaxFrameRate));

    //    if (NewFrameLength < RequestedMinFrameLengthLines)
    //        NewFrameLength = RequestedMinFrameLengthLines;
    //}

    //// Clamp to sensor's limit
    //if (NewFrameLength > pContext->MaxFrameLength)
    //    NewFrameLength = pContext->MaxFrameLength;
    //else if (NewFrameLength < pContext->MinFrameLength)
    //    NewFrameLength = pContext->MinFrameLength;
    //
    //// Need to update FrameLength?
    //if (NewFrameLength != pContext->FrameLength)
    //{
    //    NVODM_WRITE_RETURN_ON_ERROR(&pContext->I2c,
    //            0x0340, (NewFrameLength >> 8) & 0xFF);
    //    NVODM_WRITE_RETURN_ON_ERROR(&pContext->I2c,
    //            0x0341, NewFrameLength & 0xFF);
    //    pContext->FrameLength = NewFrameLength;

    //    // Update variables depending on FrameLength.
    //    pContext->FrameRate = (NvF32)pContext->VtPixClkFreqHz /
    //                          (NvF32)(pContext->FrameLength *
    //                                  pContext->LineLength);
    //}
    //
    //// FrameLength should have been updated but we have to check again
    //// in case FrameLength gets clamped.
    //if (NewCoarseTime >= pContext->FrameLength)
    //{
    //    NewCoarseTime = pContext->FrameLength - 1;
    //}
    //
    //// Current CoarseTime needs to be updated?
    //if (pContext->CoarseTime != NewCoarseTime)
    //{
    //    NVODM_WRITE_RETURN_ON_ERROR(&pContext->I2c,
    //            0x0202, (NewCoarseTime >> 8) & 0xFF);
    //    NVODM_WRITE_RETURN_ON_ERROR(&pContext->I2c,
    //            0x0203, NewCoarseTime & 0xFF);

    //    // Calculate the new exposure based on the sensor and sensor
    //    // settings.
    //    pContext->Exposure = ((NvF32)NewCoarseTime * LineLength +
    //                          SENSOR_BAYER_DEFAULT_EXPOSURE_FINE) /
    //                         Freq;
    //    pContext->CoarseTime = NewCoarseTime;
    //}

    return NV_TRUE;
}

/**
 * SensorBayer_SetInputClock. Phase 2. Sensor Dependent.
 * Setting the input clock and updating the variables based on the input clock
 * The frame rate and exposure calculations are directly related to the
 * clock speed, so this is how the camera driver (the controller of the clocks)
 * can inform the nvodm driver of the current clock speed.
 */
static NvBool
SensorBayer_SetInputClock(
    SensorBayerContext *pContext,
    const NvOdmImagerClockProfile *pClockProfile)
{

    if (pClockProfile->ExternalClockKHz <
        g_SensorBayerCaps.InitialSensorClockRateKHz)
        return NV_FALSE;

    pContext->SensorInputClockkHz = pClockProfile->ExternalClockKHz;
    pContext->SensorClockRatio = pClockProfile->ClockMultiplier;

    // if sensor is initialized (SetMode has been called), then we need to
    // update necessary sensor variables (based on this particular sensor)
    if (pContext->SensorInitialized)
    {
        pContext->VtPixClkFreqHz =
            (NvU32)((pContext->SensorInputClockkHz * 1000 * pContext->PllMult) /
                    (pContext->PllPreDiv * pContext->PllVtPixDiv *
                     pContext->PllVtSysDiv));

        pContext->Exposure = (NvF32)(pContext->CoarseTime *
                                     pContext->LineLength +
                                     SENSOR_BAYER_DEFAULT_EXPOSURE_FINE) /
                             (NvF32)pContext->VtPixClkFreqHz;
        pContext->MaxExposure = (NvF32)((SENSOR_BAYER_DEFAULT_MAX_FRAME_LENGTH - 1) *
                                        pContext->LineLength +
                                        SENSOR_BAYER_DEFAULT_EXPOSURE_FINE) /
                                (NvF32)pContext->VtPixClkFreqHz;
        pContext->MinExposure = (NvF32)SENSOR_BAYER_DEFAULT_EXPOSURE_FINE /
                                (NvF32)pContext->VtPixClkFreqHz;

    }

    return NV_TRUE;
}


/** 
 * SensorBayer_WriteGains. Phase 2. Sensor Dependent.
 * Writing the sensor registers for the new gains.
 * Just like exposure, this allows the sensor to brighten an image. If the
 * exposure time is insufficient to make the picture viewable, gains are 
 * applied.  During calibration, the gains will be measured for maximum 
 * effectiveness before the noise level becomes too high.
 * If per-channel gains are available, they are used to normalize the color.
 * Most sensors output overly greenish images, so the red and blue channels
 * are increased.
 */
static NvBool
SensorBayer_WriteGains(
    SensorBayerContext *pContext,
    const NvF32 *pGains)
{

	//WriteToI2C(GainSequence);
    memcpy(pContext->Gains, pGains, sizeof(NvF32)*4);

    return NV_TRUE;
}


BOOL InitOmnivision(void)
{
	BYTE i;

	if(GetI2CLock(500)){
		for (i=0; i<20; i++)
		{
			SetOmnivisionReg(0xff, 0x01);
			if(GetOmnivisionReg(0x1c)==0x7f && GetOmnivisionReg(0x1d) == 0xa2)
				break;
			Sleep(2);
		}

		if(GetOmnivisionReg(0x1c)!=0x7f || GetOmnivisionReg(0x1d)!=0xa2)
		{
	//		MessageBox(NULL, L"Omnivision Sensor not Found 1!", L"ERROR", MB_OK);
	//		CIFDeInit();
			ReleaseI2CLock();
			return FALSE;
		}

		// Add your register initialisation here

	//	Initialize  CIF YCbCr setting
	//
	//	60 ff 01 ;
		SetOmnivisionReg(0xff, 0x01);
	//	60 12 80 ;soft reset
		SetOmnivisionReg(0x12, 0x80);	// SOFT RESET
	//	;delay 1ms 
		Sleep(5);

	//	60 ff 00
		SetOmnivisionReg(0xff, 0x00);	// set page 0 registers ----------------------------------------
	//	60 2c ff ;added for rev2
		SetOmnivisionReg(0x2c, 0xff);	// added for REV 2
	//	60 2e df ;added for rev2
		SetOmnivisionReg(0x2e, 0xdf);	// added for REV 2


		if(GetOmnivisionReg(0x2c)!=0xff || GetOmnivisionReg(0x2e)!=0xdf)
		{
	//		MessageBox(NULL, L"Registers 0x2C,2D READ ERROR!", L"ERROR", MB_OK);
			ReleaseI2CLock();
			return FALSE;
		}

		
	//	60 ff 01 ;
		SetOmnivisionReg(0xff, 0x01);	// set page 1 registers ----------------------------------------
	//	60 3c 32 ;for Frex to GND mod @031606
		SetOmnivisionReg(0x3c, 0x32);	// for Frex to GND mod @031606
		
	//	60 11 00 ; 
		SetOmnivisionReg(0x11, 0x00);	// 
	//	60 09 02 ;mod @031606
		SetOmnivisionReg(0x09, 0x02);	// mod @031060 
	//	60 04 28 ; 
		SetOmnivisionReg(0x04, 0x28);	// 
	//	60 13 e5 ; 
		SetOmnivisionReg(0x13, 0xe5);	// 
	//	60 14 48 ;
		SetOmnivisionReg(0x14, 0x48);	// 
	//	60 2c 0c ; 
		SetOmnivisionReg(0x2c, 0x0c);	// 
	//	60 33 78 ; 
		SetOmnivisionReg(0x33, 0x78);	// 
	//	60 3a 33 ;mod @031606
		SetOmnivisionReg(0x3a, 0x33);	// mod @031606
	//	60 3b fB ;mod @031606
		SetOmnivisionReg(0x3b, 0xfb);	// mod @031606

		//	60 3e 00 ; 
		SetOmnivisionReg(0x3e, 0x00);	// 
	//	60 43 11 ;mod for rev2
		SetOmnivisionReg(0x43, 0x11);	// mod for rev2
	//	60 16 10 ;mod for rev2
		SetOmnivisionReg(0x16, 0x10);	// mod for rev2

	//	60 39 02 ;mod @031606
		SetOmnivisionReg(0x39, 0x02);	// mod @031606

	//	60 35 88 ;mod for rev2
		SetOmnivisionReg(0x35, 0x88);	// mod for rev2
	//	60 22 0a ;mod for rev2
		SetOmnivisionReg(0x22, 0x0a);	// mod for rev2
	//	60 37 40 ;mod for rev2
		SetOmnivisionReg(0x37, 0x40);	// mod for rev2
	//	60 23 00 ;add for rev2
		SetOmnivisionReg(0x23, 0x00);	// mod for rev2
	//	60 34 a0 ;add for rev2
		SetOmnivisionReg(0x34, 0xa0);	// mod for rev2
	//	60 36 1a ;
		SetOmnivisionReg(0x36, 0x1a);
	//	60 06 02 ;add for rev2
		SetOmnivisionReg(0x06, 0x02);	// mod for rev2
	//	60 07 c0 ;add for rev2
		SetOmnivisionReg(0x07, 0xc0);	// mod for rev2
	//	60 0d b7 ;mod for rev2
		SetOmnivisionReg(0x0d, 0xb7);	// mod for rev2
	//	60 0e 01 ;mod for rev2
		SetOmnivisionReg(0x0e, 0x01);	// mod for rev2
	//	60 4c 00 ;add for rev2
		SetOmnivisionReg(0x4c, 0x00);	// mod for rev2

	//	60 4a 81 ;add for rev2
		SetOmnivisionReg(0x4a, 0x81);	// mod for rev2
	//	60 21 99 ;add for rev2
		SetOmnivisionReg(0x21, 0x99);	// mod for rev2

	//	60 24 3a ;mod @100206 
		SetOmnivisionReg(0x24, 0x3a);	// mod @031606
	//	60 25 32 ;mod @100206
		SetOmnivisionReg(0x25, 0x32);	// mod @031606
	//	60 26 82 ;
		SetOmnivisionReg(0x26, 0x82);	// 
	//	60 5c 00 ;rev2 mod @030306
		SetOmnivisionReg(0x5c, 0x00);	// REV 2 mod @031606
	//	60 63 00 ;rev2 mod @030306
		SetOmnivisionReg(0x63, 0x00);	// REV 2 mod @031606
		;
	//	60 5d 55 ;add @100206
		SetOmnivisionReg(0x5d, 0x55);	// add @100206
	//	60 5e 7d ;add @100206
		SetOmnivisionReg(0x5e, 0x7d);	// add @100206
	//	60 5f 7d ;add @100206
		SetOmnivisionReg(0x5f, 0x7d);	// add @100206
	//	60 60 55 ;add @100206
		SetOmnivisionReg(0x60, 0x55);	// add @100206

	//	60 61 70 ; 
		SetOmnivisionReg(0x61, 0x70);	// 
	//	60 62 80 ; 
		SetOmnivisionReg(0x62, 0x80);	// 
	//	60 7c 05 ;mod @030306
		SetOmnivisionReg(0x7c, 0x05);	// mod @030306
		;
	//	60 20 80 ; 
		SetOmnivisionReg(0x20, 0x80);	// 
	//	60 28 30 ;
		SetOmnivisionReg(0x28, 0x30);	// 
	//	60 6c 00 ;add for rev2
		SetOmnivisionReg(0x6c, 0x00);	// add for rev2 
	//	60 6d 80 ;
		SetOmnivisionReg(0x6d, 0x80);	// 
	//	60 6e 00 ;
		SetOmnivisionReg(0x6e, 0x00);	// 
	//	60 70 02 ;mod for rev2
		SetOmnivisionReg(0x70, 0x02);	// mod for rev2 
	//	60 71 94 ;mod for rev2
		SetOmnivisionReg(0x71, 0x94);	// mod for rev2
	//	60 73 c1 ;	SetOmnivisionReg(0x28, 0x30);	// 
		SetOmnivisionReg(0x73, 0xc1);	// 
		;
	//	60 3d 34 ; 
		SetOmnivisionReg(0x3d, 0x34);	// 
	//	60 5a 57 ;add for rev2 
		SetOmnivisionReg(0x5a, 0x57);	// add for rev2 
	//	60 4f bb ; 
		SetOmnivisionReg(0x4f, 0xbb);	// 
	//	60 50 9c ;
		SetOmnivisionReg(0x50, 0x9c);	// 
		;
		;
	//	60 ff 00 ;
		SetOmnivisionReg(0xff, 0x00);	// set page 0 ---------------------------------------------- 
	//	60 e5 7f ;add for rev2  
		SetOmnivisionReg(0xe5, 0x7f);	// add for rev2 
	//	60 f9 c0 ;add for rev2  
		SetOmnivisionReg(0xf9, 0xc0);	// add for rev2 
	//	60 41 24 ;
		SetOmnivisionReg(0x41, 0x24);	//
	//	60 e0 14 ;
		SetOmnivisionReg(0xe0, 0x14);	//
	//	60 76 ff ; 
		SetOmnivisionReg(0x76, 0xff);	//
	//	60 33 a0
		SetOmnivisionReg(0x33, 0xa0);	//
	//	60 42 20
		SetOmnivisionReg(0x42, 0x20);	//
	//	60 43 18
		SetOmnivisionReg(0x43, 0x18);	//
	//	60 4c 00
		SetOmnivisionReg(0x4c, 0x00);	//
	//	60 87 d0
		SetOmnivisionReg(0x87, 0xd0);	//
	//	60 88 3f ;mod for rev2 
		SetOmnivisionReg(0x88, 0x3f);	// mod for rev2
	//	60 d7 03
		SetOmnivisionReg(0xd7, 0x03);	//
	//	60 d9 10
		SetOmnivisionReg(0xd9, 0x10);	//
	//	60 d3 82 ;
		SetOmnivisionReg(0xd3, 0x82);	//

	//	60 c8 08 ;
		SetOmnivisionReg(0xc8, 0x08);	//
	//	60 c9 80 ;
		SetOmnivisionReg(0xc9, 0x80);	//

	//	60 7c 00 ; 
		SetOmnivisionReg(0x7c, 0x00);	// SDE Indirect register ADDRESS
	//	60 7d 00 ; 
		SetOmnivisionReg(0x7d, 0x00);	// SDE Indirect Register DATA
	//	60 7c 03 ;
		SetOmnivisionReg(0x7c, 0x03);	//
	//	60 7d 48 ;
		SetOmnivisionReg(0x7d, 0x48);	//
	//	60 7d 48 ;
		SetOmnivisionReg(0x7d, 0x48);	//
	//	60 7c 08 ; 
		SetOmnivisionReg(0x7c, 0x08);	//
	//	60 7d 20 ;
		SetOmnivisionReg(0x7d, 0x20);	//
	//	60 7d 10 ;
		SetOmnivisionReg(0x7d, 0x10);	//
	//	60 7d 0e ;
		SetOmnivisionReg(0x7d, 0x0e);	//
		;
	//	60 90 00 
		SetOmnivisionReg(0x90, 0x00);	//
	//	60 91 0e
		SetOmnivisionReg(0x91, 0x0e);	//
	//	60 91 1a
		SetOmnivisionReg(0x91, 0x1a);	//
	//	60 91 31
		SetOmnivisionReg(0x91, 0x31);	//
	//	60 91 5a
		SetOmnivisionReg(0x91, 0x5a);	//
	//	60 91 69
		SetOmnivisionReg(0x91, 0x69);	//
	//	60 91 75
		SetOmnivisionReg(0x91, 0x75);	//
	//	60 91 7e
		SetOmnivisionReg(0x91, 0x7e);	//
	//	60 91 88
		SetOmnivisionReg(0x91, 0x88);	//
	//	60 91 8f
		SetOmnivisionReg(0x91, 0x8f);	//
	//	60 91 96
		SetOmnivisionReg(0x91, 0x96);	//
	//	60 91 a3
		SetOmnivisionReg(0x91, 0xa3);	//
	//	60 91 af
		SetOmnivisionReg(0x91, 0xaf);	//
	//	60 91 c4
		SetOmnivisionReg(0x91, 0xc4);	//
	//	60 91 d7
		SetOmnivisionReg(0x91, 0xd7);	//
	//	60 91 e8
		SetOmnivisionReg(0x91, 0xe8);	//
	//	60 91 20
		SetOmnivisionReg(0x91, 0x20);	//
		;
	//	60 92 00 ;
		SetOmnivisionReg(0x92, 0x00);	//
	//	60 93 06 ; 
		SetOmnivisionReg(0x93, 0x06);	//
	//	60 93 e3
		SetOmnivisionReg(0x93, 0xe3);	//
	//	60 93 05
		SetOmnivisionReg(0x93, 0x05);	//
	//	60 93 05
		SetOmnivisionReg(0x93, 0x05);	//
	//	60 93 00
		SetOmnivisionReg(0x93, 0x00);	//
	//	60 93 02 ;mod @100206
		SetOmnivisionReg(0x93, 0x02);	// mod for rev2
	//	60 93 00
		SetOmnivisionReg(0x93, 0x00);	//
	//	60 93 00
		SetOmnivisionReg(0x93, 0x00);	//
	//	60 93 00
		SetOmnivisionReg(0x93, 0x00);	//
	//	60 93 00
		SetOmnivisionReg(0x93, 0x00);	//
	//	60 93 00
		SetOmnivisionReg(0x93, 0x00);	//
	//	60 93 00
		SetOmnivisionReg(0x93, 0x00);	//
	//	60 93 00
		SetOmnivisionReg(0x93, 0x00);	//
		;
	//	60 96 00 ; 
		SetOmnivisionReg(0x96, 0x00);	//
	//	60 97 08 ; 
		SetOmnivisionReg(0x97, 0x08);	//
	//	60 97 19 ; 
		SetOmnivisionReg(0x97, 0x19);	//
	//	60 97 02 ;
		SetOmnivisionReg(0x97, 0x02);	//
	//	60 97 0c ; 
		SetOmnivisionReg(0x97, 0x0c);	//
	//	60 97 24 ; 
		SetOmnivisionReg(0x97, 0x24);	//
	//	60 97 30 ;
		SetOmnivisionReg(0x97, 0x30);	//
	//	60 97 28 ; 
		SetOmnivisionReg(0x97, 0x28);	//
	//	60 97 26 ;
		SetOmnivisionReg(0x97, 0x26);	//
	//	60 97 02 ;
		SetOmnivisionReg(0x97, 0x02);	//
	//	60 97 98 ; 
		SetOmnivisionReg(0x97, 0x98);	//
	//	60 97 80 ; 
		SetOmnivisionReg(0x97, 0x80);	//
	//	60 97 00 ;
		SetOmnivisionReg(0x97, 0x00);	//
	//	60 97 00 ;
		SetOmnivisionReg(0x97, 0x00);	//
		;
	//	60 c3 ed ; 
		SetOmnivisionReg(0xc3, 0xed);	//
	//	60 a4 00 ;
		SetOmnivisionReg(0xa4, 0x00);	//
	//	60 a8 00 ;
		SetOmnivisionReg(0xa8, 0x00);	//
	//	60 c5 11
		SetOmnivisionReg(0xc5, 0x11);	//
	//	60 c6 51 ;
		SetOmnivisionReg(0xc6, 0x51);	//
	//	60 bf 80 ;mod for rev2
		SetOmnivisionReg(0xbf, 0x80);	// mod for rev2
	//	60 ba ff ;add @101006
		SetOmnivisionReg(0xba, 0xff);	// add @101006
	//	60 bb 00 ;add @101006
		SetOmnivisionReg(0xbb, 0x00);	// add @101006
	//	60 bc 64 ;add @101006
		SetOmnivisionReg(0xbc, 0x64);	// add @101006
	//	60 c7 10 ;mod for rev2 
		SetOmnivisionReg(0xc7, 0x10);	// mod for rev2
	//	60 b6 66 ; 
		SetOmnivisionReg(0xb6, 0x66);	//
	//	60 b8 A5
		SetOmnivisionReg(0xb8, 0xa5);	//
	//	60 b7 64
		SetOmnivisionReg(0xb7, 0x64);	//
	//	60 b9 7C
		SetOmnivisionReg(0xb9, 0x7c);	//
	//	60 b3 af
		SetOmnivisionReg(0xb3, 0xaf);	//
	//	60 b4 97
		SetOmnivisionReg(0xb4, 0x97);	//
	//	60 b5 FF
		SetOmnivisionReg(0xb5, 0xff);	//
	//	60 b0 C5
		SetOmnivisionReg(0xb0, 0xc5);	//
	//	60 b1 94
		SetOmnivisionReg(0xb1, 0x94);	//
	//	60 b2 0f
		SetOmnivisionReg(0xb2, 0x0f);	//
	//	60 c4 5c ;mod for rev2 
		SetOmnivisionReg(0xc4, 0x5c);	// mod for rev2

		////////////////////////////////////////////////////////////////////

		//	60 c0 c8 ;
		SetOmnivisionReg(0xc0, 0xc8);	// HSIZE8[10:3] 
	//	60 c1 96 ;
		SetOmnivisionReg(0xc1, 0x96);	// VSIZE8[10:3]
	//	60 8c 00 ; 
		SetOmnivisionReg(0x8c, 0x00);	// 480 x 272
	//	60 86 3d ;CIF 
		SetOmnivisionReg(0x86, 0x3d);	// CTRL2 : CIF mode
	//	60 50 92
		SetOmnivisionReg(0x50, 0x89);	// CTRL7:was 92 
	//	60 51 70
		SetOmnivisionReg(0x51, 0x90);	// HSIZE was 80
	//	60 52 2c
		SetOmnivisionReg(0x52, 0xe3);	// VSIZE was 2c
	//	60 53 40
		SetOmnivisionReg(0x53, 0x00);	// XOFF: was 40
	//	60 54 00
		SetOmnivisionReg(0x54, 0x92);	// YOFF was 00
	//	60 55 88
		SetOmnivisionReg(0x55, 0x08);	// VHYX: was 88 
	//	60 57 00
	//	SetOmnivisionReg(0x57, 0x00);	// TEST: default 0
	//	60 5a 58
		SetOmnivisionReg(0x5a, 0x78);	// ZMOW: default 0x58
	//	60 5b 48
		SetOmnivisionReg(0x5b, 0x44);	// ZMOH: default 0x48
	//	60 5c 00
		SetOmnivisionReg(0x5c, 0x00);	// ZMHH: default 0

	//	60 c3 ed ; 
		SetOmnivisionReg(0xc3, 0xed);	//
	//	60 7f 00 ; 
		SetOmnivisionReg(0x7f, 0x00);	//

	//	60 da 00 ;YUV
		SetOmnivisionReg(0xda, 0x01);	// was 00 IMAGE MODE:YUV !that fixed the color problem!
	//	60 d3 04 ;
		SetOmnivisionReg(0xd3, 0x04);	//

	//	60 e5 1f ;
		SetOmnivisionReg(0xe5, 0x1f);	// 
	//	60 e1 67 ;mod 060606 
		SetOmnivisionReg(0xe1, 0x67);	// mod 060606
	//	60 e0 00 ; 
		SetOmnivisionReg(0xe0, 0x00);	// all resets off
	//	60 dd 7f ; 
		SetOmnivisionReg(0xdd, 0x7f);	//
	//	60 05 00 ; 
		SetOmnivisionReg(0x05, 0x00);	// R_BYPASS: DSP NOT BYPASSED 

										
		SetOmnivisionReg(0xff, 0x01);

		if(GetOmnivisionReg(0x1c)!=0x7f || GetOmnivisionReg(0x1d)!=0xa2)
		{
	//		MessageBox(NULL, L"Omnivision Sensor not Found 2!", L"ERROR", MB_OK);
			ReleaseI2CLock();
			return FALSE;
		}
	}
ReleaseI2CLock();	
return TRUE;
}


/**
 * SensorBayer_Open. Phase 1.
 * Initialize sensor bayer and its private context
 */
static NvBool SensorBayer_Open(NvOdmImagerHandle hImager)
{
    SensorBayerContext *pSensorBayerContext = NULL;

    if (!hImager || !hImager->pSensor)
        return NV_FALSE;

    pSensorBayerContext =
        (SensorBayerContext*)malloc(sizeof(SensorBayerContext));
    if (!pSensorBayerContext)
        goto fail;

    memset(pSensorBayerContext, 0, sizeof(SensorBayerContext));
    
	//Do Hardreset and Enable Sensor
	//SetPinAltFn(SODIMM_PIN_SENSOR_ENABLE,-1,DIR_OUT);
	SetPinAltFn(SODIMM_PIN_SENSOR_RESET,-1,DIR_OUT);
	//SetPinLevel(SODIMM_PIN_SENSOR_ENABLE,0);
	SetPinLevel(SODIMM_PIN_SENSOR_RESET,0); // assert reset on sensor
	Sleep(100);
	SetPinLevel(SODIMM_PIN_SENSOR_RESET,1); // deassert reset on sensor

	if(!I2CInit())
		return FALSE;

	InitOmnivision();

    pSensorBayerContext->NumModes =
        NV_ARRAY_SIZE(g_SensorBayerSetModeSequenceList);
    pSensorBayerContext->ModeIndex =
        pSensorBayerContext->NumModes + 1; // invalid mode
    
    /** 
     * Phase 2. Initialize exposure and gains.
     */
    pSensorBayerContext->Exposure = -1.0; // invalid exposure

     // convert gain's unit to float
    pSensorBayerContext->MaxGain = SENSOR_GAIN_TO_F32(SENSOR_BAYER_DEFAULT_MAX_GAIN);
    pSensorBayerContext->MinGain = SENSOR_GAIN_TO_F32(SENSOR_BAYER_DEFAULT_MIN_GAIN);
    
    /**
     * Phase 2 ends here.
     */

    hImager->pSensor->pPrivateContext = pSensorBayerContext;

    return NV_TRUE;    

fail:
    free(pSensorBayerContext);
    return NV_FALSE;
}


/**
 * SensorBayer_Close. Phase 1.
 * Free the sensor's private context
 */
static void SensorBayer_Close(NvOdmImagerHandle hImager)
{
    SensorBayerContext *pContext = NULL;
        
    if (!hImager || !hImager->pSensor || !hImager->pSensor->pPrivateContext)
        return;

    pContext = (SensorBayerContext*)hImager->pSensor->pPrivateContext;

    // cleanup
    free(pContext);
    hImager->pSensor->pPrivateContext = NULL;
}

/**
 * SensorBayer_GetCapabilities. Phase 1.
 * Returnt sensor bayer's capabilities
 */
static void
SensorBayer_GetCapabilities(
    NvOdmImagerHandle hImager,
    NvOdmImagerCapabilities *pCapabilities)
{
    // Copy the sensor caps from g_SensorBayerCaps
    memcpy(pCapabilities, &g_SensorBayerCaps,
        sizeof(NvOdmImagerCapabilities));
}

/**
 * SensorBayer_ListModes. Phase 1.
 * Return a list of modes that sensor bayer supports.
 * If called with a NULL ptr to pModes, then it just returns the count
 */
static void
SensorBayer_ListModes(
    NvOdmImagerHandle hImager,
    NvOdmImagerSensorMode *pModes,
    NvS32 *pNumberOfModes)
{
    SensorBayerContext *pContext =
        (SensorBayerContext*)hImager->pSensor->pPrivateContext;
    NvS32 i;

    if (pNumberOfModes)
    {
        *pNumberOfModes = (NvS32)pContext->NumModes;
        if (pModes)
        {
            // Copy the modes from g_SensorBayerSetModeSequenceList
            for (i = 0; i < *pNumberOfModes; i++)
                pModes[i] = g_SensorBayerSetModeSequenceList[i].Mode;
        }
    }
}

static NV_INLINE NvBool SensorBayerChooseModeIndex(
    SensorBayerContext *pContext,
    NvSize Resolution,
    NvU32 *pIndex)
{
    NvU32 Index;

    for (Index = 0; Index < pContext->NumModes; Index++)
    {
        if ((Resolution.width ==
             g_SensorBayerSetModeSequenceList[Index].Mode.
             ActiveDimensions.width) &&
            (Resolution.height ==
             g_SensorBayerSetModeSequenceList[Index].Mode.
             ActiveDimensions.height))
        {
            *pIndex = Index;
            return NV_TRUE;
        }
    }

    return NV_FALSE;
}

/**
 * SensorBayer_SetMode. Phase 1.
 * Set sensor bayer to the mode of the desired resolution and frame rate.
 */
static NvBool
SensorBayer_SetMode(
    NvOdmImagerHandle hImager,
    const SetModeParameters *pParameters,
    NvOdmImagerSensorMode *pSelectedMode,
    SetModeParameters *pResult)
{
    NvBool Status = NV_FALSE;
    SensorBayerContext *pContext =
        (SensorBayerContext*)hImager->pSensor->pPrivateContext;
    NvU32 Index;
    ModeDependentSettings *pModeSettings;

    Status = SensorBayerChooseModeIndex(pContext, pParameters->Resolution,
                                        &Index);

    // No match found
    if (!Status)
        return NV_FALSE;

    if (pSelectedMode)
    {
        *pSelectedMode = g_SensorBayerSetModeSequenceList[Index].Mode;
    }

    if (pContext->ModeIndex == Index)
        return NV_TRUE;

    // TODO: Do we need to set a lower coarse time for new framelenghtline?
    // I don't think so because the set mode sequence reset the sensor.

    // i2c writes for the set mode sequence
    //Status = WriteI2CSequence(&pContext->I2c,
    //             g_SensorBayerSetModeSequenceList[Index].pSequence,
    //             pContext->SensorInitialized && pContext->FastSetMode);
	WriteToI2C(g_SensorBayerSetModeSequenceList[Index].pSequence);

    if (!Status)
        return NV_FALSE;
    /**
     * the following is Phase 2.
     */
    pContext->SensorInitialized = NV_TRUE;

    // Update sensor context after set mode
    //NV_ASSERT(pContext->SensorInputClockkHz > 0);

    // These hardcoded numbers are from the set mode sequence and this formula
    // is based on this sensor. Sensor Dependent.
    pContext->PllMult = SENSOR_BAYER_DEFAULT_PLL_MULT;
    pContext->PllPreDiv = SENSOR_BAYER_DEFAULT_PLL_PRE_DIV;
    pContext->PllVtPixDiv = SENSOR_BAYER_DEFAULT_PLL_VT_PIX_DIV;
    pContext->PllVtSysDiv = SENSOR_BAYER_DEFAULT_PLL_VT_SYS_DIV;
    pContext->VtPixClkFreqHz =
        (NvU32)((pContext->SensorInputClockkHz * 1000 * pContext->PllMult) /
                (pContext->PllPreDiv * pContext->PllVtPixDiv *
                 pContext->PllVtSysDiv));

    pModeSettings =
        (ModeDependentSettings*)g_SensorBayerSetModeSequenceList[Index].
                                pModeDependentSettings;

    pContext->LineLength = pModeSettings->LineLength;
    pContext->FrameLength = pModeSettings->FrameLength;
    pContext->CoarseTime = pModeSettings->CoarseTime;

    pContext->Exposure = (NvF32)(pContext->CoarseTime * pContext->LineLength +
                                 SENSOR_BAYER_DEFAULT_EXPOSURE_FINE) /
                         (NvF32)pContext->VtPixClkFreqHz;
    pContext->MaxExposure = (NvF32)((SENSOR_BAYER_DEFAULT_MAX_FRAME_LENGTH - 1) *
                                    pContext->LineLength +
                                    SENSOR_BAYER_DEFAULT_EXPOSURE_FINE) /
                            (NvF32)pContext->VtPixClkFreqHz;
    pContext->MinExposure = (NvF32)SENSOR_BAYER_DEFAULT_EXPOSURE_FINE /
                            (NvF32)pContext->VtPixClkFreqHz;

    pContext->MaxFrameLength = SENSOR_BAYER_DEFAULT_MAX_FRAME_LENGTH;
    pContext->MinFrameLength = pModeSettings->MinFrameLength;

	pContext->FrameRate = 30;
	pContext->MaxFrameRate = 0;
	pContext->MinFrameRate = 120;
	
	//DAN
    //pContext->FrameRate = (NvF32)pContext->VtPixClkFreqHz /
    //                      (NvF32)(pContext->FrameLength *
    //                              pContext->LineLength); 
    //pContext->MaxFrameRate = pContext->FrameRate;
    //pContext->MinFrameRate = (NvF32)pContext->VtPixClkFreqHz /
    //                         (NvF32)(SENSOR_BAYER_DEFAULT_MAX_FRAME_LENGTH *
    //                                 pContext->LineLength);

    pContext->Gains[0] = SENSOR_BAYER_DEFAULT_GAIN_R;
    pContext->Gains[1] = SENSOR_BAYER_DEFAULT_GAIN_GR;
    pContext->Gains[2] = SENSOR_BAYER_DEFAULT_GAIN_GB;
    pContext->Gains[3] = SENSOR_BAYER_DEFAULT_GAIN_B;



    SensorBayer_WriteExposure(pContext, &pParameters->Exposure);
    SensorBayer_WriteGains(pContext, pParameters->Gains);

    // Do we need to slow down the frame rate?
    //if (pContext->RequestedMaxFrameRate > 0.0f)
    //{
    //    NvU32 RequestedMinFrameLengthLines = 
    //        (NvU32)(pContext->VtPixClkFreqHz / (pContext->LineLength *
    //                pContext->RequestedMaxFrameRate));

    //    if (RequestedMinFrameLengthLines > pContext->MaxFrameLength)
    //        RequestedMinFrameLengthLines = pContext->MaxFrameLength;

    //    if (pContext->FrameLength < RequestedMinFrameLengthLines)
    //    {
    //        NVODM_WRITE_RETURN_ON_ERROR(&pContext->I2c,
    //                0x0340, (RequestedMinFrameLengthLines >> 8) & 0xFF);
    //        NVODM_WRITE_RETURN_ON_ERROR(&pContext->I2c,
    //                0x0341, RequestedMinFrameLengthLines & 0xFF);
    //        
    //        pContext->FrameLength = RequestedMinFrameLengthLines;
    //    }
    //}


    if (pResult)
    {
        pResult->Resolution = g_SensorBayerSetModeSequenceList[Index].
                              Mode.ActiveDimensions;
        pResult->Exposure = pContext->Exposure;
        memcpy(pResult->Gains, &pContext->Gains, sizeof(NvF32) * 4);
    }

    /**
     * Phase 2 ends here.
     */

    pContext->ModeIndex = Index;

    //Wait 2 frames for gains/exposure to take effect.
    Sleep((NvU32)(2000.0 / (NvF32)(pContext->FrameRate)));

    return Status;
}


/**
 * SensorBayer_SetPowerLevel. Phase 1
 * Set the sensor's power level
 */
static NvBool
SensorBayer_SetPowerLevel(
    NvOdmImagerHandle hImager,
    NvOdmImagerPowerLevel PowerLevel)
{
    NvBool Status = NV_TRUE;

    SensorBayerContext *pContext =
        (SensorBayerContext*)hImager->pSensor->pPrivateContext;
    //const NvOdmPeripheralConnectivity *pConnections;

    if (pContext->PowerLevel == PowerLevel)
        return NV_TRUE;

    ////NV_ASSERT(hImager->pSensor->pConnections);
    //pConnections = hImager->pSensor->pConnections;

    switch(PowerLevel)
    {
        case NvOdmImagerPowerLevel_On:
 /*           Status = SetPowerLevelWithPeripheralConnectivity(pConnections,
                         &pContext->I2c, &pContext->GpioConfig,
                         NvOdmImagerPowerLevel_On);
            if (!Status) {
                return NV_FALSE;
            }*/

            break;

        case NvOdmImagerPowerLevel_Standby:

            //Status = SetPowerLevelWithPeripheralConnectivity(pConnections,
            //             &pContext->I2c, &pContext->GpioConfig,
            //             NvOdmImagerPowerLevel_Standby);

            //break;
        case NvOdmImagerPowerLevel_Off:

            // // Set the sensor to standby mode
            //NVODM_WRITE_RETURN_ON_ERROR(&pContext->I2c, 0x0100, 0);

            //Status = SetPowerLevelWithPeripheralConnectivity(pConnections,
            //             &pContext->I2c, &pContext->GpioConfig,
            //             NvOdmImagerPowerLevel_Off);
            //
            //// wait for the standby 0.1 sec should be enough
            //NvOdmOsWaitUS(100000); 

            break;
        default:
            //NV_ASSERT("!Unknown power level\n");
            Status = NV_FALSE;
            break;
    }

    if (Status)
        pContext->PowerLevel = PowerLevel;

    return Status;
}

/**
 * SensorBayer_SetParameter. Phase 2.
 * Set sensor specific parameters. 
 * For Phase 1. This can return NV_TRUE.
 */
static NvBool
SensorBayer_SetParameter(
    NvOdmImagerHandle hImager,
    NvOdmImagerParameter Param,
    NvS32 SizeOfValue,
    const void *pValue)
{
    NvBool Status = NV_TRUE;
    SensorBayerContext *pContext =
        (SensorBayerContext*)hImager->pSensor->pPrivateContext;

    switch(Param)
    {
        // Phase 2.
        case NvOdmImagerParameter_SensorExposure:
			if(SizeOfValue!=sizeof(NvF32)){
				NKDbgPrintfW(L"ERROR: SensorBayer_SetParameter (SensorExposure): SizeOfValue does not match \n\r");
				return FALSE;
			}

            Status = SensorBayer_WriteExposure(pContext, (NvF32*)pValue);
            break;

        // Phase 2.
        case NvOdmImagerParameter_SensorGain:
			if(SizeOfValue!=4*sizeof(NvF32)){
				NKDbgPrintfW(L"ERROR: SensorBayer_SetParameter (SensorGain): SizeOfValue does not match \n\r");
				return FALSE;
			}

            Status = SensorBayer_WriteGains(pContext, pValue);
            break;

        // Phase 2.
        case NvOdmImagerParameter_SensorInputClock:
			if(SizeOfValue!=sizeof(NvOdmImagerClockProfile)){
				NKDbgPrintfW(L"ERROR: SensorBayer_SetParameter (SensorInputClock): SizeOfValue does not match \n\r");
				return FALSE;
			}
            
            Status = SensorBayer_SetInputClock(
                      pContext,
                      ((NvOdmImagerClockProfile*)pValue));
            break;

        // Phase 3.
        case NvOdmImagerParameter_OptimizeResolutionChange:
			//if(SizeOfValue!=sizeof(BOOL)){
			//	NKDbgPrintfW(L"ERROR: SensorBayer_SetParameter (OptimizeResolutionChange): SizeOfValue does not match \n\r");
			//	return FALSE;
			//}

            pContext->FastSetMode = *((NvBool*)pValue);
            break;

        // This is optional but nice to have.
        //case NvOdmImagerParameter_SelfTest:
        //    //TODO: Not Implemented.
        //    break;
        // Phase 3.
        case NvOdmImagerParameter_Reset:
            return NV_FALSE;
            break;

        case NvOdmImagerParameter_MaxSensorFrameRate:
			if(SizeOfValue!=sizeof(NvF32)){
				NKDbgPrintfW(L"ERROR: SensorBayer_SetParameter (MaxSensorFrameRate): SizeOfValue does not match \n\r");
				return FALSE;
			}

            pContext->RequestedMaxFrameRate = *((NvF32*)pValue);
            break;

        default:
            NKDbgPrintfW(L"ERROR: SensorBayer_SetParameter Parameter not implemented \n\r");
            break;
    }
    
    return NV_TRUE;
}

/**
 * SensorBayer_GetParameter. Phase 1.
 * Get sensor specific parameters
 */
static NvBool
SensorBayer_GetParameter(
    NvOdmImagerHandle hImager,
    NvOdmImagerParameter Param,
    NvS32 SizeOfValue,
    void *pValue)
{
	 NvBool Status = NV_TRUE; 
    NvF32 *pFloatValue = pValue;
    SensorBayerContext *pContext =
        (SensorBayerContext*)hImager->pSensor->pPrivateContext;
    NvU32 *pUIntValue = pValue;

    switch(Param)
    {
        // Phase 1.
        case NvOdmImagerParameter_CalibrationData:
 /*           CHECK_PARAM_SIZE_RETURN_MISMATCH(SizeOfValue,
                sizeof(NvOdmImagerCalibrationData));*/
            {
                NvOdmImagerCalibrationData *pCalibration =
                        (NvOdmImagerCalibrationData*)pValue;
                pCalibration->CalibrationData = pSensorCalibrationData;
                pCalibration->NeedsFreeing = NV_FALSE;
            }
            break;

        // Phase 1, it can return min = max = 1.0f
        // Phase 2, return the real numbers.
        case NvOdmImagerParameter_SensorExposureLimits:
            //CHECK_PARAM_SIZE_RETURN_MISMATCH(SizeOfValue, 2 * sizeof(NvF32));

            pFloatValue[0] = pContext->MinExposure;
            pFloatValue[1] = pContext->MaxExposure;
            break;

        // Phase 1, it can return min = max = 1.0f
        // Phase 2, return the real numbers.
        case NvOdmImagerParameter_SensorGainLimits:
            //CHECK_PARAM_SIZE_RETURN_MISMATCH(SizeOfValue, 2 * sizeof(NvF32));
            pFloatValue[0] = pContext->MinGain;
            pFloatValue[1] = pContext->MaxGain;

            break;

        // Get the sensor status. This is optional but nice to have.
        case NvOdmImagerParameter_DeviceStatus:
            //CHECK_PARAM_SIZE_RETURN_MISMATCH(SizeOfValue,
            //    sizeof(NvOdmImagerDeviceStatus));
            {
                NvOdmImagerDeviceStatus *pStatus;
                pStatus = (NvOdmImagerDeviceStatus *)pValue;
                {
                    //NvOdmImagerI2cConnection *pI2c = &pContext->I2c;
                    // Pick your own useful registers to use for debug
                    // If the camera hangs, these register values are printed
                    // Sensor Dependent.
/*
                    NvOdmImagerI2cRead(pI2c, 0x0000, &pStatus->Values[0]);
                    NvOdmImagerI2cRead(pI2c, 0x303B, &pStatus->Values[1]);
                    NvOdmImagerI2cRead(pI2c, 0x303C, &pStatus->Values[2]);
                    NvOdmImagerI2cRead(pI2c, 0x306A, &pStatus->Values[3]);
*/
                    pStatus->Count = 4;
                }
            }
            break;
        case NvOdmImagerParameter_SensorFrameRate:
            //CHECK_PARAM_SIZE_RETURN_MISMATCH(SizeOfValue, sizeof(NvF32));
            {
                pFloatValue[0] = 30.00; // TODO update with a calculated rate
            }
            Status = NV_TRUE;
            break;

        case NvOdmImagerParameter_SensorFrameRateLimits:
            //CHECK_PARAM_SIZE_RETURN_MISMATCH(SizeOfValue,
            //    2 * sizeof(NvF32));
            pFloatValue[0] = pContext->MinFrameRate;
            pFloatValue[1] = pContext->MaxFrameRate;
            break;

        case NvOdmImagerParameter_RegionUsedByCurrentResolution:
            if (SizeOfValue != sizeof(NvOdmImagerRegion)) {
                //NV_ASSERT(!"Bad size");
                return NV_FALSE;
            }
            if (pContext->ModeIndex == (pContext->NumModes+1)) { // Invalid mode
                // No info available until we've actually set a mode.
                return NV_FALSE;
            }
            {
                NvOdmImagerRegion *pRegion = (NvOdmImagerRegion *)pValue;
                pRegion->RegionStart.x = 0;
                pRegion->RegionStart.y = 0;
                if (pContext->ModeIndex == 0) { // Quarter res
                    pRegion->xScale = 2;
                    pRegion->yScale = 2;
                } else { // Should be Full res.
                    pRegion->xScale = 1;
                    pRegion->yScale = 1;
                }
                return NV_TRUE;
            }
            break;

        case NvOdmImagerParameter_SensorFrameRateLimitsAtResolution:
            //CHECK_PARAM_SIZE_RETURN_MISMATCH(SizeOfValue,
            //    sizeof(NvOdmImagerFrameRateLimitAtResolution));
            {
                NvOdmImagerFrameRateLimitAtResolution *pData;
                NvU32 Index;
                NvBool MatchFound = NV_FALSE;
                ModeDependentSettings *pModeSettings = NULL;

                pData = (NvOdmImagerFrameRateLimitAtResolution *)pValue;
                MatchFound =
                    SensorBayerChooseModeIndex(pContext, pData->Resolution,
                    &Index);

                if (!MatchFound)
                {
                    pData->MinFrameRate = 0.0f;
                    pData->MaxFrameRate = 0.0f;
                    break;
                }

                pModeSettings = (ModeDependentSettings*)
                    g_SensorBayerSetModeSequenceList[Index].
                    pModeDependentSettings;

				pData->MaxFrameRate = 30.0;
				pData->MinFrameRate = 0.0;
				//DAN
                //pData->MaxFrameRate = (NvF32)pContext->VtPixClkFreqHz /
                //                      (NvF32)(pModeSettings->FrameLength *
                //                              pModeSettings->LineLength);
                //pData->MinFrameRate =
                //    (NvF32)pContext->VtPixClkFreqHz /
                //    (NvF32)(SENSOR_BAYER_DEFAULT_MAX_FRAME_LENGTH *
                //            pModeSettings->LineLength);

            }
            break;

        // Phase 3.
        // Get the override config data.
        case NvOdmImagerParameter_CalibrationOverrides:
			//DAN
			NKDbgPrintfW(L"DAN TO DO NvOdmImagerParameter_CalibrationOverrides\n\r");
			/*
            CHECK_PARAM_SIZE_RETURN_MISMATCH(SizeOfValue,
                sizeof(NvOdmImagerCalibrationData));
            {
                NvOdmImagerCalibrationData *pCalibration =
                    (NvOdmImagerCalibrationData*)pValue;

                pCalibration->CalibrationData =
                    LoadOverridesFile(pOverrideFiles,
                        (sizeof(pOverrideFiles)/sizeof(pOverrideFiles[0])));
                pCalibration->NeedsFreeing = (pCalibration->CalibrationData != NULL);
                Status = pCalibration->NeedsFreeing;
            }
			*/
            break;

        case NvOdmImagerParameter_SensorExposureLatchTime:
            //CHECK_PARAM_SIZE_RETURN_MISMATCH(SizeOfValue,
            //    sizeof(NvU32));
                *pUIntValue = 1;
            break;

        // Phase 3.
        default:
            //NV_ASSERT("!Not Implemented");
            Status = NV_FALSE;
    }
 
    return Status;
}

/**
 * SensorBayer_GetPowerLevel. Phase 3.
 * Get the sensor's current power level
 */
static void
SensorBayer_GetPowerLevel(
    NvOdmImagerHandle hImager,
    NvOdmImagerPowerLevel *pPowerLevel)
{
    SensorBayerContext *pContext =
        (SensorBayerContext*)hImager->pSensor->pPrivateContext;

    *pPowerLevel = pContext->PowerLevel;
}

/**
 * SensorBayer_GetHal. Phase 1.
 * return the hal functions associated with sensor bayer
 */
NvBool SensorPlugin_GetHal_Bayer_OV2640(NvOdmImagerHandle hImager)
{
    if (!hImager || !hImager->pSensor)
        return NV_FALSE;

    hImager->pSensor->pfnOpen = SensorBayer_Open;
    hImager->pSensor->pfnClose = SensorBayer_Close;
    hImager->pSensor->pfnGetCapabilities = SensorBayer_GetCapabilities;
    hImager->pSensor->pfnListModes = SensorBayer_ListModes;
    hImager->pSensor->pfnSetMode = SensorBayer_SetMode;
    hImager->pSensor->pfnSetPowerLevel = SensorBayer_SetPowerLevel;
    hImager->pSensor->pfnGetPowerLevel = SensorBayer_GetPowerLevel;
    hImager->pSensor->pfnSetParameter = SensorBayer_SetParameter;
    hImager->pSensor->pfnGetParameter = SensorBayer_GetParameter;

    return NV_TRUE;
}
