/*
 * Copyright (c) 2008 NVIDIA Corporation.  All rights reserved.
 *
 * NVIDIA Corporation and its licensors retain all intellectual property
 * and proprietary rights in and to this software, related documentation
 * and any modifications thereto.  Any use, reproduction, disclosure or
 * distribution of this software and related documentation without an express
 * license agreement from NVIDIA Corporation is strictly prohibited.
 */

#include "sensor_common.h"
#include "nvodm_imager.h"
#include "nvodm_imager_int.h"
#include "imager_hal.h" //shard ressources with driver
#include "I2CLib.h"
#include "GPIOLib.h"
#include "CoProcLib.h"

/****************************************************************************	
 * Implementing for a New Sensor:
 * 
 * When implementing for a new sensor from this template, there are four 
 * phases:
 * Phase 0: Gather sensor information. In phase 0, we should gather the sensor
 *          information for bringup such as sensor capabilities, set mode i2c
 *          write sequence, NvOdmIoAddress in nvodm_query_discovery_imager.h,
 *          and so on.
 * Phase 1: Bring up a sensor. After phase 1, we should be able to do still
 *          captures. Since this is a yuv sensor, there won't be many
 *          interactions between camera and sensor.
 * Phase 2: Fully functioning sensor. After phase 2, the sensor should be fully
 *          functioning as described in nvodm_imager.h
 *
 * To help implementing for a new sensor, the template code will be marked as
 * Phase 0, 1, 2
 ****************************************************************************/



/**
 * ----------------------------------------------------------
 *  Start of Phase 0 code
 * ----------------------------------------------------------
 */

/**
 * Sensor Specific Variables/Defines. Phase 0.
 *
 * If a new sensor code is created from this template, the following
 * variable/defines should be created for the new sensor specifically.
 * 1. A sensor capabilities (NvOdmImagerCapabilities) for this specific sensor
 * 2. A sensor set mode sequence list  (an array of SensorSetModeSequence)
 *    consisting of pairs of a sensor mode and corresponding set mode sequence
 * 3. A set mode sequence for each mode defined in 2.
 */

/**
 * Sensor yuv's capabilities. Phase 0. Sensor Dependent.
 * This is the first thing the camera driver requests. The information here
 * is used to setup the proper interface code (CSI, VIP, pixel type), start
 * up the clocks at the proper speeds, and so forth.
 * For Phase 0, one could start with these below as an example for a yuv
 * sensor and just update clocks from what is given in the reference documents
 * for that device. The FAE may have specified the clocks when giving the
 * reccommended i2c settings.
 */

static NvOdmImagerCapabilities g_SensorYuvCaps =
{
    "ADC7280",  // string identifier, used for debugging
    NvOdmImagerSensorInterface_Parallel8, // interconnect type
    {
		//ITU_R BT.656 (default) stream contains 4:2:2 data with optional embedded synchronization codes
        NvOdmImagerPixelType_UYVY, //NvOdmImagerPixelType_VYUY,//NvOdmImagerPixelType_UYVY,
    },
    NvOdmImagerOrientation_0_Degrees, // rotation with respect to main display
    NvOdmImagerDirection_Away,          // toward or away the main display
    27000, // initial sensor clock (in khz)
	//VGA (640 x 480): 15 fps at 12 MHz (default), programmable up to 30 fps at 27 MHz
    { {27000, 3.0}, {0,0} }, // preferred clock profile
    {
        NvOdmImagerSyncEdge_Falling, // hsync polarity
        NvOdmImagerSyncEdge_Falling, // vsync polarity
        NV_FALSE,                   // uses mclk on vgp0
    }, // parallel settings (VIP)
    {0,0,0,NV_FALSE,0}, // serial settings (CSI) (not used)
    {0, 0}, // min blank time, shouldn't need adjustment
    0, // preferred mode, which resolution to use on startup
    0, // FOCUSER_GUID, only set if focuser code available
    0, // FLASH_GUID, only set if flash device code is available
    NVODM_IMAGER_CAPABILITIES_END
};

#define I2C_ADDRESS			0x21
#define VPP_SLAVE_ADDRESS	0x42

/* User Sub Map Regs */
#define ADV7280_INPUT_CONTROL 0x00
#define ADV7280_VIDEO_SELECTION_1 0x01
#define ADV7280_VIDEO_SELECTION_2 0x02
#define ADV7280_OUTPUT_CONTROL 0x03
#define ADV7280_EXTENDED_OUTPUT_CONTROL 0x04
#define ADV7280_AUTODETECT_ENABLE 0x07
#define ADV7280_ADI_CONTROL_1 0x0E
#define ADV7280_POWER_MANAGEMENT 0x0F

#define ADV7280_STATUS_1 0x10
#define ADV7280_STATUS_2 0x12
#define ADV7280_STATUS_3 0x13

#define ADV7280_IDENT 0x11
#define ADV7280_SHAPING_FILTER_CONTROL_1 0x17
#define ADV7280_ADI_CONTROL_2 0x1D
#define ADV7280_PIXEL_DELAY_CONTROL 0x27
#define ADV7280_VPP_SLAVE_ADDRESS 0xFD
#define ADV7280_OUTPUT_SYNC_SELECT_2 0x6B

/* VPP regs */
#define VPP_DEINT_RESET 0x41
#define VPP_I2C_DEINT_ENABLE 0x55
#define VPP_ADV_TIMING_MODE_EN 0x5B

#define ADV7280_EXTENDED_OUTPUT_CONTROL_NTSCDIS		0x07
#define ADV7280_AUTODETECT_DEFAULT			0x7f

#define ADV7280_INPUT_CONTROL_COMPOSITE_IN1		0x00
#define ADV7280_INPUT_CONTROL_AD_PAL_BG_NTSC_J_SECAM	0x00

#define ADV7280_INPUT_CONTROL_NTSC_M 0x50
#define ADV7280_INPUT_CONTROL_PAL60 0x60
#define ADV7280_INPUT_CONTROL_NTSC_443 0x70
#define ADV7280_INPUT_CONTROL_PAL_BG 0x80
#define ADV7280_INPUT_CONTROL_PAL_N 0x90
#define ADV7280_INPUT_CONTROL_PAL_M 0xa0
#define ADV7280_INPUT_CONTROL_PAL_M_PED 0xb0
#define ADV7280_INPUT_CONTROL_PAL_COMB_N 0xc0
#define ADV7280_INPUT_CONTROL_PAL_COMB_N_PED 0xd0
#define ADV7280_INPUT_CONTROL_PAL_SECAM 0xe0

#define ADV7280_ADI_CTRL_IRQ_SPACE			0x20

#define ADV7280_STATUS1_IN_LOCK		0x01
#define ADV7280_STATUS1_AUTOD_MASK	0x70
#define ADV7280_STATUS1_AUTOD_NTSM_M_J	0x00
#define ADV7280_STATUS1_AUTOD_NTSC_4_43 0x10
#define ADV7280_STATUS1_AUTOD_PAL_M	0x20
#define ADV7280_STATUS1_AUTOD_PAL_60	0x30
#define ADV7280_STATUS1_AUTOD_PAL_B_G	0x40
#define ADV7280_STATUS1_AUTOD_SECAM	0x50
#define ADV7280_STATUS1_AUTOD_PAL_COMB	0x60
#define ADV7280_STATUS1_AUTOD_SECAM_525	0x70

#define ADV7280_ICONF1_ADI		0x40
#define ADV7280_ICONF1_ACTIVE_LOW	0x01
#define ADV7280_ICONF1_PSYNC_ONLY	0x10

#define ADV7280_IMR1_ADI	0x44
#define ADV7280_IMR2_ADI	0x48
#define ADV7280_IRQ3_AD_CHANGE	0x08
#define ADV7280_ISR3_ADI	0x4A
#define ADV7280_ICR3_ADI	0x4B
#define ADV7280_IMR3_ADI	0x4C
#define ADV7280_IMR4_ADI	0x50

static DevCtrlReg16 SetModeSequence_800x624_YUV_25FPS[] =
{
    {ADV7280_POWER_MANAGEMENT, 0x80}, //Reset Chip
    {SEQUENCE_WAIT_MS, 10},
	{ADV7280_POWER_MANAGEMENT, 0x00}, //Initialize and Exit power down mode
	{ADV7280_ADI_CONTROL_1,0x80}, //analog devices recommends
    {0x9C, 0x00}, //analog device recommends
    {0x9C, 0xFF},
    {ADV7280_ADI_CONTROL_1, 0x00}, // Enter User Sub Map
    {0x14, 0x11}, //ANALOG_CLAMP_CONTROL: Select 100% color bars Test pattern
    {ADV7280_OUTPUT_CONTROL, 0x0C}, //Enable Pixel & Sync output drivers
    {ADV7280_EXTENDED_OUTPUT_CONTROL, ADV7280_EXTENDED_OUTPUT_CONTROL_NTSCDIS}, //User space
    {ADV7280_SHAPING_FILTER_CONTROL_1, 0x41}, //Enable SH1 01 linux
    {0x39, 0x24}, //Disable comb filtering
    {ADV7280_ADI_CONTROL_2, 0x40}, //Enable LLC output driver
    {ADV7280_OUTPUT_SYNC_SELECT_2, 0x01}, //VSYNC on VS/FIELD/SFL pin
    {ADV7280_INPUT_CONTROL, ADV7280_INPUT_CONTROL_AD_PAL_BG_NTSC_J_SECAM | ADV7280_INPUT_CONTROL_COMPOSITE_IN1}, 
    {ADV7280_AUTODETECT_ENABLE, ADV7280_AUTODETECT_DEFAULT}, //Enable autodetection
    {0x52, 0xCD}, //analog devices recommends
    {0x80, 0x51},
    {0x81, 0x51},
    {0x82, 0x68},
    {0x0E, 0x80},
    {0xD9, 0x44},
    {0x0E, 0x00},
    {ADV7280_VPP_SLAVE_ADDRESS, VPP_SLAVE_ADDRESS << 1}, // Set VPP Map
    {SEQUENCE_CHANGE_ADDR, VPP_SLAVE_ADDRESS}, // Set I2C Address to VPP
    {0xA3, 0x00}, //VPP - not documented
    {VPP_ADV_TIMING_MODE_EN, 0x03}, //VPP - Enbable Advanced Timing Mode
    {VPP_I2C_DEINT_ENABLE, 0x80}, //VPP - Enable Deinterlacer 
    {SEQUENCE_CHANGE_ADDR, I2C_ADDRESS}, // Set I2C Address back to normal
    {SEQUENCE_END, 0},
};

/**
 * A list of sensor mode and its SetModeSequence. Phase 0. Sensor Dependent.
 * When SetMode is called, it searches the list for desired resolution and
 * frame rate and write the SetModeSequence through i2c.
 */
static SensorSetModeSequence g_SensorYuvSetModeSequenceList[] =
{
    // NOTE: Width and Height must be a multiple of 16 !
    // {{Width x Height},			 {x_skip, y_skip}, fps,  AR, PLL Mult}, Set Mode Sequence
	//{{{720, (240)}, {0,0},  30, 0}, NULL, NULL},
    {{{720, 576}, {68, 38},  50.0, 1.0, 3.0}, SetModeSequence_800x624_YUV_25FPS, NULL}, //858x625
	//{{{720, 288}, {0, 0},  30, 0}, NULL, NULL},
};

/**
 * ----------------------------------------------------------
 *  Start of Phase 1 code
 * ----------------------------------------------------------
 */

DWORD WINAPI TimeoutThreadProc();

static BOOL	g_bSensorInit = FALSE;

#if 1
//------------------------------------------------------------------------------------
/// Try to access all Slave Addresses without data. If a slave is present, it will ACK the address
static void i2cScan(void)
{
    int i, count=0;
    BOOL Success;
	BYTE data[10];

    wprintf (L"Scanning i2c bus...\r\n");
    for (i = 1; i < 127; i++)
    {
        Success = I2CBurstRead(i, data, 0, 1); //I2CBurstWrite(i, NULL, 0, 0);
        if (Success)
        {
            wprintf(L"%s0x%02x", ((count++ == 0)? L"  " : L", "), i);
        }
        //Sleep(1);
    }
    wprintf (L"\r\n%d device(s) found.\r\n", count);
}



static DWORD WriteADV7280(WORD offset, WORD data)
{
	return SensorI2C_Write(offset, data);
}

static DWORD ReadADV7280(BYTE offset, WORD* pData)
{
	return SensorI2C_Read(offset, pData);
}


static BOOL I2CSetup()
{
	BOOL Success;
	BYTE data[20];
	BOOL ret = FALSE;
	DWORD VerMaj;
    DWORD VerMin;
    DWORD VerBuild;
	//UINT32 i;

	I2cIntGetLibVersion(&VerMaj, &VerMin, &VerBuild);
    //wprintf(L"Demo Application for I2cLib\r\n");
    //wprintf(L"Library Version %d.%d Build %d\r\n\r\n", VerMaj, VerMin, VerBuild);
		
	if(getModuleFamily() == 1)
	{
		ret = I2CInitEx(0x7000C500);
	}
	else if(getModuleFamily() == 0)
	{
		ret = I2CInit();
	}
	
	if(ret)		
	{
		 
		I2CSetSpeed(I2C_400KBPS);
		
		Success = GetI2CLock(500);	

		//data[0]=0xA0;  //RESET
		//ret = I2CBurstWrite(I2C_ADDRESS, data, ADV7280_POWER_MANAGEMENT, 1);		

		Sleep(10);

		data[0]=0x00;  //Initialize and Exit power down mode
		if(I2CBurstWrite(I2C_ADDRESS, data, ADV7280_POWER_MANAGEMENT, 1) == FALSE) goto ERROR_EXIT;
		
		
		data[0]=0x80;  //analog devices recommends 
		if(I2CBurstWrite(I2C_ADDRESS, data, ADV7280_ADI_CONTROL_1, 1) == FALSE) goto ERROR_EXIT;

		data[0]=0x00;  //analog device recommends
		if(I2CBurstWrite(I2C_ADDRESS, data, 0x9C, 1)  == FALSE) goto ERROR_EXIT;

		data[0]=0xFF;  //analog device recommends
		if(I2CBurstWrite(I2C_ADDRESS, data, 0x9C, 1)  == FALSE) goto ERROR_EXIT;

		data[0]=0x00;  // Enter User Sub Map 
		if(I2CBurstWrite(I2C_ADDRESS, data, ADV7280_ADI_CONTROL_1, 1)  == FALSE) goto ERROR_EXIT;
		
		data[0]=0x11;  // Enter User Sub Map 100% color bar 
		if(I2CBurstWrite(I2C_ADDRESS, data, 0x14, 1)  == FALSE) goto ERROR_EXIT;

		//Enable Pixel & Sync output drivers 
		data[0]=0x0C; 
		if(I2CBurstWrite(I2C_ADDRESS, data, ADV7280_OUTPUT_CONTROL, 1)  == FALSE) goto ERROR_EXIT;
		
		data[0]=ADV7280_EXTENDED_OUTPUT_CONTROL_NTSCDIS;  //User space
		if(I2CBurstWrite(I2C_ADDRESS, data, ADV7280_EXTENDED_OUTPUT_CONTROL, 1)  == FALSE) goto ERROR_EXIT;

		data[0]=0x41; //Enable SH1 01 linux
		if(I2CBurstWrite(I2C_ADDRESS, data, ADV7280_SHAPING_FILTER_CONTROL_1, 1) == FALSE) goto ERROR_EXIT;
		
		//Disable comb filtering
		data[0]=0x24;
		if(I2CBurstWrite(I2C_ADDRESS, data, 0x39, 1)  == FALSE) goto ERROR_EXIT;
		
		//Enable LLC output driver
		data[0]=0x40;  
		if(I2CBurstWrite(I2C_ADDRESS, data, ADV7280_ADI_CONTROL_2, 1)  == FALSE) goto ERROR_EXIT;

		//VSYNC on VS/FIELD/SFL pin
		data[0]=0x01;  
		if(I2CBurstWrite(I2C_ADDRESS, data, ADV7280_OUTPUT_SYNC_SELECT_2, 1)  == FALSE) goto ERROR_EXIT;
		
		//Enable autodetection 
		data[0]=ADV7280_INPUT_CONTROL_AD_PAL_BG_NTSC_J_SECAM | ADV7280_INPUT_CONTROL_COMPOSITE_IN1;  //Hidden space
		if(I2CBurstWrite(I2C_ADDRESS, data, ADV7280_INPUT_CONTROL, 1)  == FALSE) goto ERROR_EXIT;

		data[0]=ADV7280_AUTODETECT_DEFAULT;
		if(I2CBurstWrite(I2C_ADDRESS, data, ADV7280_AUTODETECT_ENABLE, 1)  == FALSE) goto ERROR_EXIT;
		
		//analog devices recommends
		data[0]=0xCD;  
		if(I2CBurstWrite(I2C_ADDRESS, data, 0x52, 1)  == FALSE) goto ERROR_EXIT;
		
		//analog devices recommends
		data[0]=0x51;  
		if(I2CBurstWrite(I2C_ADDRESS, data, 0x80, 1)  == FALSE) goto ERROR_EXIT;
		
		// analog devices recommends
		data[0]=0x51;  
		if(I2CBurstWrite(I2C_ADDRESS, data, 0x81, 1)  == FALSE) goto ERROR_EXIT;

		// analog devices recommends
		data[0]=0x68;  
		if(I2CBurstWrite(I2C_ADDRESS, data, 0x82, 1)  == FALSE) goto ERROR_EXIT;
		
		data[0]=0x80;  
		if(I2CBurstWrite(I2C_ADDRESS, data, 0x0E, 1)  == FALSE) goto ERROR_EXIT;
		
		data[0]=0x44;  
		if(I2CBurstWrite(I2C_ADDRESS, data, 0xD9, 1)  == FALSE) goto ERROR_EXIT;
		
		data[0]=0x00;  
		if(I2CBurstWrite(I2C_ADDRESS, data, 0x0E, 1)  == FALSE) goto ERROR_EXIT;

		// Set VPP Map 
		data[0]=VPP_SLAVE_ADDRESS << 1;  
		if(I2CBurstWrite(I2C_ADDRESS, data, ADV7280_VPP_SLAVE_ADDRESS, 1)  == FALSE) goto ERROR_EXIT;
		
		//VPP - not documented 
		data[0]=0x00;  
		if(I2CBurstWrite(VPP_SLAVE_ADDRESS, data, 0xA3, 1)  == FALSE) goto ERROR_EXIT;
		
		//VPP - Enbable Advanced Timing Mode
		data[0]=0x03;  
		if(I2CBurstWrite(VPP_SLAVE_ADDRESS, data, VPP_ADV_TIMING_MODE_EN, 1)  == FALSE) goto ERROR_EXIT;
		
		//VPP - Enable Deinterlacer 
		data[0]=0x80;  
		if(I2CBurstWrite(VPP_SLAVE_ADDRESS, data, VPP_I2C_DEINT_ENABLE, 1)  == FALSE) goto ERROR_EXIT; 
	
		/* NKDbgPrintfW(L"\t00\t01\t02\t03\t04\t05\t06\t07\t08\t09\t0a\t0b\t0c\t0d\t0e\t0f\n");
		
		for(i=0; i<255; i+=16)
		{
			I2CBurstRead(I2C_ADDRESS,data,i,16);
			NKDbgPrintfW(L"%x:\t%2x\t%2x\t%2x\t%2x\t%2x\t%2x\t%2x\t%2x\t%2x\t%2x\t%2x\t%2x\t%2x\t%2x\t%2x\t%2x\r\n",
				i,data[0],data[1],data[2],data[3],data[4],data[5],data[6],data[7],data[8],data[9],data[10],data[11],data[12],data[13],data[14],data[15]);
		}  */
		//CreateThread(NULL, 0, TimeoutThreadProc, 0, 0, NULL);
		TimeoutThreadProc();
		
		Sleep(500);

		ReleaseI2CLock();
		g_bSensorInit = TRUE;	
	}
	
	return 1;

ERROR_EXIT:	
	NKDbgPrintfW(L"Sensor Init Failed");
	ReleaseI2CLock();
	I2CDeInit();
	return 0 ;
}

DWORD count = 200; //in terms of 0.1 second count

// Status Register Read thread
static DWORD WINAPI TimeoutThreadProc()
{		
	BYTE read[20];
	BOOL ret;

	while(count)
	{		
		GetI2CLock(500);

		/* I2CBurstRead(I2C_ADDRESS,read,ADV7280_STATUS_1,1);
		NKDbgPrintfW(L"ADV7280_STATUS_1:\t%2x\r\n",read[0]);
	
		read[0]=0x00;  
		ret = I2CBurstRead(I2C_ADDRESS, read, ADV7280_STATUS_2,1);
		NKDbgPrintfW(L"ADV7280_STATUS_2:\t%2x\r\n",read[0]); */

		read[0]=0x00; 
		ret = I2CBurstRead(I2C_ADDRESS, read, ADV7280_STATUS_3,1);
		NKDbgPrintfW(L"ADV7280_STATUS_3:\t%2x\r\n",read[0]);
		
		if(read[0] & 0x40) //Wait for Interlaced Detection.
			break;

		ReleaseI2CLock();
		
		count--;
		
		Sleep(100);		
	}
	
	I2CDeInit();
	return 0;

}
#endif

/**
 * Sensor yuv's private context. Phase 1.
 */
typedef struct SensorYuvContextRec
{
    NvU32 ModeIndex;
    NvU32 NumModes;
    NvOdmImagerPowerLevel PowerLevel;
    NvF32 FrameRate;
}SensorYuvContext;

/**
 * Static functions
 * Some compilers require these to be declared upfront.
 */
static NvBool SensorYuv_Open(NvOdmImagerHandle hImager);
static void SensorYuv_Close(NvOdmImagerHandle hImager);

static void
SensorYuv_GetCapabilities(
    NvOdmImagerHandle hImager,
    NvOdmImagerCapabilities *pCapabilities);

static void
SensorYuv_ListModes(
    NvOdmImagerHandle hImager,
    NvOdmImagerSensorMode *pModes,
    NvS32 *pNumberOfModes);
    
static NvBool
SensorYuv_SetMode(
    NvOdmImagerHandle hImager,
    const SetModeParameters *pParameters,
    NvOdmImagerSensorMode *pSelectedMode,
    SetModeParameters *pResult);

static NvBool
SensorYuv_SetPowerLevel(
    NvOdmImagerHandle hImager,
    NvOdmImagerPowerLevel PowerLevel);

static NvBool
SensorYuv_SetParameter(
    NvOdmImagerHandle hImager,
    NvOdmImagerParameter Param,
    NvS32 SizeOfValue,
    const void *pValue);

static NvBool
SensorYuv_GetParameter(
    NvOdmImagerHandle hImager,
    NvOdmImagerParameter Param,
    NvS32 SizeOfValue,
    void *pValue);

static void
SensorYuv_GetPowerLevel(
    NvOdmImagerHandle hImager,
    NvOdmImagerPowerLevel *pPowerLevel);

/**
 * SensorYuv_Open. Phase 1.
 * Initialize sensor yuv and its private context
 */
static NvBool SensorYuv_Open(NvOdmImagerHandle hImager)
{
    SensorYuvContext *pSensorYuvContext = NULL;

    if (!hImager || !hImager->pSensor)
        return NV_FALSE;

    pSensorYuvContext =
        (SensorYuvContext*)malloc(sizeof(SensorYuvContext));
    if (!pSensorYuvContext)
        goto fail;

    memset(pSensorYuvContext, 0, sizeof(SensorYuvContext));

	DEBUGMSG(ZONE_FUNC, (_T("+SensorYuv_Open\n")));
 
    pSensorYuvContext->NumModes =
        NV_ARRAY_SIZE(g_SensorYuvSetModeSequenceList);
    pSensorYuvContext->ModeIndex =
        pSensorYuvContext->NumModes + 1; // invalid mode
    
    hImager->pSensor->pPrivateContext = pSensorYuvContext;

    g_SensorI2C_Addr = I2C_ADDRESS;
    g_SensorI2C_Speed = I2C_400KBPS;
    g_SensorI2C_OffsetWidth = 1;
    g_SensorI2C_DataWidth = 1;

    DEBUGMSG(ZONE_FUNC, (_T("-SensorYuv_Open\n")));

    return NV_TRUE;    

fail:
    free(pSensorYuvContext);
    return NV_FALSE;
}

/**
 * SensorYuv_Close. Phase 1.
 * Free the sensor's private context
 */
static void SensorYuv_Close(NvOdmImagerHandle hImager)
{
    SensorYuvContext *pContext = NULL;
        
    if (!hImager || !hImager->pSensor || !hImager->pSensor->pPrivateContext)
        return;

    pContext = (SensorYuvContext*)hImager->pSensor->pPrivateContext;
    
    DEBUGMSG(ZONE_FUNC, (_T("+SensorYuv_Close\n")));

    // cleanup
    free(pContext);
    hImager->pSensor->pPrivateContext = NULL;

    DEBUGMSG(ZONE_FUNC, (_T("+SensorYuv_Close\n")));
}

/**
 * SensorYuv_GetCapabilities. Phase 1.
 * Returnt sensor Yuv's capabilities
 */
static void
SensorYuv_GetCapabilities(
    NvOdmImagerHandle hImager,
    NvOdmImagerCapabilities *pCapabilities)
{
    DEBUGMSG(ZONE_FUNC, (_T("+SensorYuv_GetCapabilities\n")));
	// Copy the sensor caps from g_SensorYuvCaps
    memcpy(pCapabilities, &g_SensorYuvCaps,
        sizeof(NvOdmImagerCapabilities));
    DEBUGMSG(ZONE_FUNC, (_T("-SensorYuv_GetCapabilities\n")));
}

/**
 * SensorYuv_ListModes. Phase 1.
 * Return a list of modes that sensor yuv supports.
 * If called with a NULL ptr to pModes, then it just returns the count
 */
static void
SensorYuv_ListModes(
    NvOdmImagerHandle hImager,
    NvOdmImagerSensorMode *pModes,
    NvS32 *pNumberOfModes)
{
    SensorYuvContext *pContext =
        (SensorYuvContext*)hImager->pSensor->pPrivateContext;
    NvS32 i;

    DEBUGMSG(ZONE_FUNC, (_T("+SensorYuv_ListModes\n")));

    if (pNumberOfModes)
    {
        *pNumberOfModes = (NvS32)pContext->NumModes;
        if (pModes)
        {
            // Copy the modes from g_SensorYuvSetModeSequenceList
            for (i = 0; i < *pNumberOfModes; i++)
                pModes[i] = g_SensorYuvSetModeSequenceList[i].Mode;
        }
    }
    DEBUGMSG(ZONE_FUNC, (_T("-SensorYuv_ListModes\n")));
}

/**
 * SensorYuv_SetMode. Phase 1.
 * Set sensor yuv to the mode of the desired resolution and frame rate.
 */
static NvBool
SensorYuv_SetMode(
    NvOdmImagerHandle hImager,
    const SetModeParameters *pParameters,
    NvOdmImagerSensorMode *pSelectedMode,
    SetModeParameters *pResult)
{
    NvBool Status = NV_FALSE;
    SensorYuvContext *pContext =
        (SensorYuvContext*)hImager->pSensor->pPrivateContext;
    NvU32 Index;

    DEBUGMSG(ZONE_FUNC, (_T("+SensorYuv_SetMode\n")));
	
	// Find the right entrys in g_SensorYuvSetModeSequenceList that matches
    // the desired resolution and framerate
    for (Index = 0; Index < pContext->NumModes; Index++)
    {
        if ((pParameters->Resolution.width == g_SensorYuvSetModeSequenceList[Index].
             Mode.ActiveDimensions.width) &&
            (pParameters->Resolution.height == g_SensorYuvSetModeSequenceList[Index].
             Mode.ActiveDimensions.height))
            break;
    }

    // No match found
    if (Index == pContext->NumModes)
    {
        DEBUGMSG(ZONE_ERR, (_T("-SensorYuv_SetMode no match found\n")));
        return NV_FALSE;
    }

    if (pSelectedMode)
    {
        DEBUGMSG(ZONE_INFO, (_T("Setting resolution: %dx%d with %d fps.\n"), g_SensorYuvSetModeSequenceList[Index].Mode.ActiveDimensions.width, g_SensorYuvSetModeSequenceList[Index].Mode.ActiveDimensions.height, (NvU32)g_SensorYuvSetModeSequenceList[Index].Mode.PeakFrameRate));
        *pSelectedMode = g_SensorYuvSetModeSequenceList[Index].Mode;
    }

    if (pContext->ModeIndex == Index)
        return NV_TRUE;

    // i2c writes for the set mode sequence
	DEBUGMSG(ZONE_INFO, (_T(" SensorYuv_SetMode %d Write ModeSequence\n"), Index));
    Status = SensorI2C_WriteSequence(g_SensorYuvSetModeSequenceList[Index].pSequence);
    {
        WORD data;
        WORD offset=0x13;
        SensorI2C_Open();
        SensorI2C_Read(offset, &data);
        SensorI2C_Close();
    }

	if (!Status) {
		DEBUGMSG(ZONE_FUNC, (_T("-SensorYuv_SetMode OV5642 I2C failed\n")));
       return NV_FALSE;
	}
    //I2CSetup();

    pContext->ModeIndex = Index;

    if (pResult)
    {
        pResult->Resolution = g_SensorYuvSetModeSequenceList[Index].
                              Mode.ActiveDimensions;
        pResult->Exposure = 0;
        memset(pResult->Gains, 0, sizeof(NvF32) * 4);
    }

    DEBUGMSG(ZONE_FUNC, (_T("-SensorYuv_SetMode\n")));
    return Status;
}

/**
 * SensorYuv_SetPowerLevel. Phase 1
 * Set the sensor's power level
 */
static NvBool
SensorYuv_SetPowerLevel(
    NvOdmImagerHandle hImager,
    NvOdmImagerPowerLevel PowerLevel)
{
    NvBool Status = NV_TRUE;
    SensorYuvContext *pContext =
        (SensorYuvContext*)hImager->pSensor->pPrivateContext;

    DEBUGMSG(ZONE_FUNC, (_T("+SensorYuv_SetPowerLevel\n")));
   
    if (pContext->PowerLevel == PowerLevel)
        return NV_TRUE;

    switch(PowerLevel)
    {
        case NvOdmImagerPowerLevel_On:				
/*
            Status = SetPowerLevelWithPeripheralConnectivity(pConnections,
                         &pContext->I2c, &pContext->GpioConfig,
                         NvOdmImagerPowerLevel_On);
            if (!Status)
            {
                NvOdmOsDebugPrintf("NvOdmImagerPowerLevel_On failed.\n");
                return NV_FALSE;
            }
*/
			//I2CSetup(); //Initialize Sensor through I2C here	
            break;

        case NvOdmImagerPowerLevel_Standby:
/*
            Status = SetPowerLevelWithPeripheralConnectivity(pConnections,
                         &pContext->I2c, &pContext->GpioConfig,
                         NvOdmImagerPowerLevel_Standby);
*/
            break;
        case NvOdmImagerPowerLevel_Off:
/*
            Status = SetPowerLevelWithPeripheralConnectivity(pConnections,
                         &pContext->I2c, &pContext->GpioConfig,
                         NvOdmImagerPowerLevel_Off);
            
            // wait for the standby 0.1 sec should be enough
            NvOdmOsWaitUS(100000); 
*/
            break;
        default:
//            //NV_ASSERT("!Unknown power level\n");
            Status = NV_FALSE;
            break;
    }

    if (Status)
        pContext->PowerLevel = PowerLevel;
    
    DEBUGMSG(ZONE_FUNC, (_T("-SensorYuv_SetPowerLevel\n")));

    return Status;
}

/**
 * SensorYuv_SetParameter. Phase 2.
 * Set sensor specific parameters. 
 * For Phase 1. This can return NV_TRUE.
 */
static NvBool
SensorYuv_SetParameter(
    NvOdmImagerHandle hImager,
    NvOdmImagerParameter Param,
    NvS32 SizeOfValue,
    const void *pValue)
{
    if(Param==NvOdmImagerParameter_SensorTriggerStill)
        return NV_TRUE;
	
    DEBUGMSG(ZONE_FUNC, (_T("+SensorYuv_SetParameter(%S)\n"), GetParamName(Param)));

    switch(Param)
    {
        case NvOdmImagerParameter_SensorInputClock:
            if(SizeOfValue == sizeof(NvOdmImagerClockProfile))
            {
                // Verify the input clock is valid.
                NvU32 InClock =
                    ((NvOdmImagerClockProfile*)pValue)->ExternalClockKHz;

				if (!InClock) {
					return NV_FALSE;
				}
            }
            break;
		case NvOdmImagerParameter_Reset:		
			/*if(g_bSensorInit == FALSE)
			{				
				I2CSetup(); //Initialize Sensor through I2C here		
			}*/
			break;

        case NvOdmImagerParameter_OptimizeResolutionChange:
			if(SizeOfValue!=sizeof(NvBool)){break;}

            // Sensor yuv doesn't support optimized resolution change.
			if (*((NvBool*)pValue) == NV_TRUE)
                return NV_TRUE;
            else
                return NV_FALSE;

            break;

        case NvOdmImagerParameter_SelfTest:
            // not implemented.
            break;
        default:
            //NV_ASSERT(!"Not Implemented");
            break;
    }

    DEBUGMSG(ZONE_FUNC, (_T("-SensorYuv_SetParameter\n")));
    return NV_TRUE;
}

/**
 * SensorYuv_GetParameter. Phase 1.
 * Get sensor specific parameters
 */
static NvBool
SensorYuv_GetParameter(
    NvOdmImagerHandle hImager,
    NvOdmImagerParameter Param,
    NvS32 SizeOfValue,
    void *pValue)
{
    NvBool result = NV_TRUE;

    SensorYuvContext *pContext =
        (SensorYuvContext*)hImager->pSensor->pPrivateContext;

#if DEBUG //NvOdmImagerParameter_SensorFrameRate is called too often, output it only once
    {
        BOOL bDebugEnable = TRUE;
        static NvOdmImagerParameter OldParam = -1;

        if(Param==NvOdmImagerParameter_SensorFrameRate &&
            OldParam==NvOdmImagerParameter_SensorFrameRate )
            bDebugEnable=FALSE;
        OldParam = Param;
        if(!bDebugEnable)
            return TRUE;
    }
#endif

    DEBUGMSG(ZONE_FUNC, (L"+SensorYuv_GetParameter(%S)\n", GetParamName(Param)));

    switch(Param)
    {
        case NvOdmImagerParameter_DeviceStatus:
            if(SizeOfValue ==sizeof(NvOdmImagerDeviceStatus))
            {
                NvOdmImagerDeviceStatus *pStatus;
                pStatus = (NvOdmImagerDeviceStatus *)pValue;
            }
            break;

        case NvOdmImagerParameter_SensorFrameRate:
            if(SizeOfValue == sizeof(NvF32)){
                *((NvF32*)pValue) = g_SensorYuvSetModeSequenceList[pContext->ModeIndex].Mode.PeakFrameRate;
                break;
            }

        case NvOdmImagerParameter_GetBestSensorMode:
            if(SizeOfValue == sizeof(NvOdmImagerSensorMode))
            {
                NvOdmImagerSensorMode* pMode = (NvOdmImagerSensorMode *) pValue;
                DEBUGMSG(ZONE_INFO, (L"ModeToSet: %dx%d (%d fps)\r\n", pMode->ActiveDimensions.width, pMode->ActiveDimensions.height, (NvU32)pMode->PeakFrameRate));
                pContext->FrameRate = pMode->PeakFrameRate;
                result = NV_FALSE; // pretend we didn't process this, so Camera framework will choose the best resolution
                break;
            }

        default:
           //NV_ASSERT(!"Not Implemented");
           DEBUGMSG(ZONE_FUNC && ZONE_WARN, (L"Not Implemented\n"));
           result = NV_FALSE;
           break;
    }

    DEBUGMSG(ZONE_FUNC, (L"-SensorYuv_GetParameter\n"));

    return result;
}

/**
 * SensorYuv_GetPowerLevel. Phase 2.
 * Get the sensor's current power level
 */
static void
SensorYuv_GetPowerLevel(
    NvOdmImagerHandle hImager,
    NvOdmImagerPowerLevel *pPowerLevel)
{
    SensorYuvContext *pContext =
        (SensorYuvContext*)hImager->pSensor->pPrivateContext;

    DEBUGMSG(ZONE_FUNC, (L"+SensorYuv_GetPowerLevel\n"));

    *pPowerLevel = pContext->PowerLevel;

    DEBUGMSG(ZONE_FUNC, (L"-SensorYuv_GetPowerLevel\n"));
}

/**
 * SensorYuv_GetHal. Phase 1.
 * return the hal functions associated with sensor yuv
 */
NvBool SensorPlugin_GetHal_YUV_ADV7280(NvOdmImagerHandle hImager)
{
    if (!hImager || !hImager->pSensor)
        return NV_FALSE;

    hImager->pSensor->pfnOpen = SensorYuv_Open;
    hImager->pSensor->pfnClose = SensorYuv_Close;
    hImager->pSensor->pfnGetCapabilities = SensorYuv_GetCapabilities;
    hImager->pSensor->pfnListModes = SensorYuv_ListModes;
    hImager->pSensor->pfnSetMode = SensorYuv_SetMode;
    hImager->pSensor->pfnSetPowerLevel = SensorYuv_SetPowerLevel;
    hImager->pSensor->pfnGetPowerLevel = SensorYuv_GetPowerLevel;
    hImager->pSensor->pfnSetParameter = SensorYuv_SetParameter;
    hImager->pSensor->pfnGetParameter = SensorYuv_GetParameter;

    return NV_TRUE;
}

