/*
 * Copyright (c) 2008 NVIDIA Corporation.  All rights reserved.
 *
 * NVIDIA Corporation and its licensors retain all intellectual property
 * and proprietary rights in and to this software, related documentation
 * and any modifications thereto.  Any use, reproduction, disclosure or
 * distribution of this software and related documentation without an express
 * license agreement from NVIDIA Corporation is strictly prohibited.
 */

// ..\..\..\..\libs\lib\ARMV4IDbg

//#include "sensor_plugin.h"
//#include "nvassert.h" 
//#include "imager_util.h"
#include "sensor_common.h"
#include "nvodm_imager.h"
//#include "nvodm_imager_device.h"
#include "nvodm_imager_int.h"
#include "imager_hal.h" //shard ressources with driver
//#include "imager_util.h"

#include "GPIOLib.h"
#include "I2CLib.h"

//Not used
//#define CAM_PWRDWN_PIN				168
//#define CAM_RESET_PIN				170

//Not used
//#define CAM_RESET_POL				1
//#define CAM_PWRDWN_POL				0



#define I2C_ADDRESS						(0x78 >> 1)

/****************************************************************************
 * Implementing for a New Sensor:
 * 
 * When implementing for a new sensor from this template, there are four 
 * phases:
 * Phase 0: Gather sensor information. In phase 0, we should gather the sensor
 *          information for bringup such as sensor capabilities, set mode i2c
 *          write sequence, NvOdmIoAddress in nvodm_query_discovery_imager.h,
 *          and so on.
 * Phase 1: Bring up a sensor. After phase 1, we should be able to do still
 *          captures. Since this is a yuv sensor, there won't be many
 *          interactions between camera and sensor.
 * Phase 2: Fully functioning sensor. After phase 2, the sensor should be fully
 *          functioning as described in nvodm_imager.h
 *
 * To help implementing for a new sensor, the template code will be marked as
 * Phase 0, 1, 2
 ****************************************************************************/

/*
Callback Sequence:

Sensor_Open()
Sensor_GetCapabilities()
Sensor_ListModes()
Sensor_SetParameter(NvOdmImagerParameter_SensorInputClock)
Sensor_GetParameter(NvOdmImagerParameter_StereoCapable)
Sensor_GetParameter(NvOdmImagerParameter_SensorFrameRateLimits)
Sensor_GetParameter(NvOdmImagerParameter_GetBestSensorMode) NOTE: Return NV_FALSE to let it be handled by Camera Upper layers
Sensor_ListModes()
Sensor_SetParameter(NvOdmImagerParameter_OperationalMode)
Clock set [SENSOR]
Clock set [CORE]
Sensor_SetPowerLevel(NvOdmImagerPowerLevel_On)
Sensor_SetParameter(NvOdmImagerParameter_SensorInputClock)
Sensor_GetParameter(NvOdmImagerParameter_FlashCapabilities)
Sensor_GetParameter(NvOdmImagerParameter_TorchCapabilities)
Clock set [CORE] 24000 [External for Pads] [Internal Clock for Core] Result: 24000 KHz
Sensor_SetMode()

*/

/**
 * ----------------------------------------------------------
 *  Start of Phase 0 code
 * ----------------------------------------------------------
 */

/**
 * Sensor Specific Variables/Defines. Phase 0.
 *
 * If a new sensor code is created from this template, the following
 * variable/defines should be created for the new sensor specifically.
 * 1. A sensor capabilities (NvOdmImagerCapabilities) for this specific sensor
 * 2. A sensor set mode sequence list  (an array of SensorSetModeSequence)
 *    consisting of pairs of a sensor mode and corresponding set mode sequence
 * 3. A set mode sequence for each mode defined in 2.
 */

/**
 * Sensor yuv's capabilities. Phase 0. Sensor Dependent.
 * This is the first thing the camera driver requests. The information here
 * is used to setup the proper interface code (CSI, VIP, pixel type), start
 * up the clocks at the proper speeds, and so forth.
 * For Phase 0, one could start with these below as an example for a yuv
 * sensor and just update clocks from what is given in the reference documents
 * for that device. The FAE may have specified the clocks when giving the
 * reccommended i2c settings.
 */
static NvOdmImagerCapabilities g_SensorYuvCaps =
{
    "OV5640_Par",  // string identifier, used for debugging
    NvOdmImagerSensorInterface_Parallel8, // interconnect type
    {
        NvOdmImagerPixelType_UYVY, //NvOdmImagerPixelType_UYVY,
    },
    NvOdmImagerOrientation_0_Degrees, // rotation with respect to main display
    NvOdmImagerDirection_Away,          // toward or away the main display
    24000, // initial sensor clock (in khz)
    { {24000, 96.0f/24.0f}, {0,0} }, // preferred clock profile
    {
        NvOdmImagerSyncEdge_Rising, // hsync polarity
        NvOdmImagerSyncEdge_Rising, // vsync polarity
		NV_TRUE,                   // uses mclk on vgp0
    }, // parallel settings (VIP)
    {0,0,0,NV_FALSE,0}, // serial settings (CSI) (not used)
    {0, 0}, // min blank time, shouldn't need adjustment
    0, // preferred mode, which resolution to use on startup
    0, // FOCUSER_GUID, only set if focuser code available
    0, // FLASH_GUID, only set if flash device code is available
    NVODM_IMAGER_CAPABILITIES_END
};

#include "sensor_yuv_ov5640_regs_parallel.h"

/**
 * A list of sensor mode and its SetModeSequence. Phase 0. Sensor Dependent.
 * When SetMode is called, it searches the list for desired resolution and
 * frame rate and write the SetModeSequence through i2c.
 */
static SensorSetModeSequence g_SensorYuvSetModeSequenceList[] =
{
	// H,W muss 16 aligned sein
    // WxH				 x, y    fps    AR   PLL Mult     Set Mode Sequence
	{ { {  320,   240}, {0, 0},  30.00, 1.0, 0} , mode_320x240	, NULL},
	{ { {  640,   480}, {0, 0},  15.00, 1.0, 96.0/24.0} , SetModeSequence_VGA_YUV_15FPS/*SetModeSequence_VGA30FPS*/		, NULL		},
	{ { {  1280,   720}, {0, 0},  15.00, 1.0, 48.0/24.0} , SetModeSequence_720p_YUV, NULL		},
	{ { { 2592,  1944}, {0, 0},   7.50, 1.0, 144.0/24.0} , SetModeSequence_5MP_YUV, NULL		},
	{ { { 1600,  1200}, {0, 0},   7.50, 1.0, 144.0/24.0} , SetModeSequence_UXGA_YUV_7_5FPS, NULL		},
	{ { { 1920,  1088}, {0, 0},   15, 1.0, 96.0/24.0} , SetModeSequence_1080p_YUV, NULL	},
};
//{ { { 1920,  1088}, {0, 0},   15, 1.0, 96.0/24.0} , SetModeSequence_1080p_YUV, NULL	},

static BOOL SensorSetup()
{
//	BOOL Success;

	DEBUGMSG(ZONE_FUNC, (L"+SensorSetup\n"));
	// PINs fr OV5642 setzen
	//SetPinAltFn(CAM_PWRDWN_PIN, -1, DIR_OUT);
	//SetPinAltFn(CAM_RESET_PIN, -1, DIR_OUT);

	// AltFn fuer die anderen beiden PIN's setzen
	//SetPinAltFn(166, -1, DIR_OUT);
	//SetPinAltFn(172, -1, DIR_OUT);

	//SetPinLevel(CAM_PWRDWN_PIN, ~CAM_PWRDWN_POL);
	//Sleep(150);
	//SetPinLevel(CAM_PWRDWN_PIN, CAM_PWRDWN_POL);

	//SetPinLevel(CAM_RESET_PIN, ~CAM_RESET_POL);
	//Sleep(150);
	//SetPinLevel(CAM_RESET_PIN, CAM_RESET_POL);

	DEBUGMSG(ZONE_FUNC, (L"-SensorSetup\n"));

	return 1 ;
}


/**
 * Sensor yuv's private context. Phase 1.
 */
typedef struct SensorYuvContextRec
{
    // Phase 1 variables.
    //NvOdmImagerI2cConnection I2c;
    NvU32 ModeIndex;
    NvU32 NumModes;
    NvOdmImagerPowerLevel PowerLevel;
    NvF32 FrameRate;
    //ImagerGpioConfig GpioConfig;

}SensorYuvContext;

/**
 * Static functions
 * Some compilers require these to be declared upfront.
 */
static NvBool SensorYuv_Open(NvOdmImagerHandle hImager);
static void SensorYuv_Close(NvOdmImagerHandle hImager);

static void
SensorYuv_GetCapabilities(
    NvOdmImagerHandle hImager,
    NvOdmImagerCapabilities *pCapabilities);

static void
SensorYuv_ListModes(
    NvOdmImagerHandle hImager,
    NvOdmImagerSensorMode *pModes,
    NvS32 *pNumberOfModes);
    
static NvBool
SensorYuv_SetMode(
    NvOdmImagerHandle hImager,
    const SetModeParameters *pParameters,
    NvOdmImagerSensorMode *pSelectedMode,
    SetModeParameters *pResult);

static NvBool
SensorYuv_SetPowerLevel(
    NvOdmImagerHandle hImager,
    NvOdmImagerPowerLevel PowerLevel);

static NvBool
SensorYuv_SetParameter(
    NvOdmImagerHandle hImager,
    NvOdmImagerParameter Param,
    NvS32 SizeOfValue,
    const void *pValue);

static NvBool
SensorYuv_GetParameter(
    NvOdmImagerHandle hImager,
    NvOdmImagerParameter Param,
    NvS32 SizeOfValue,
    void *pValue);

static void
SensorYuv_GetPowerLevel(
    NvOdmImagerHandle hImager,
    NvOdmImagerPowerLevel *pPowerLevel);


/**
 * SensorYuv_Open. Phase 1.
 * Initialize sensor yuv and its private context
 */
static NvBool SensorYuv_Open(NvOdmImagerHandle hImager)
{
    SensorYuvContext *pSensorYuvContext = NULL;
	NvBool Status = NV_FALSE;

	DEBUGMSG(ZONE_FUNC, (_T("+SensorYuv_Open\n")));

	if (!hImager || !hImager->pSensor) {
		DEBUGMSG(ZONE_FUNC, (_T("-SensorYuv_Open !hImager\n")));
        return NV_FALSE;
	}

    pSensorYuvContext =
        (SensorYuvContext*)malloc(sizeof(SensorYuvContext));
    if (!pSensorYuvContext)
        goto fail;

    memset(pSensorYuvContext, 0, sizeof(SensorYuvContext));
	
    pSensorYuvContext->NumModes = NV_ARRAY_SIZE(g_SensorYuvSetModeSequenceList);
    pSensorYuvContext->ModeIndex = pSensorYuvContext->NumModes + 1; // invalid mode
    
    hImager->pSensor->pPrivateContext = pSensorYuvContext;

	//DEBUGMSG(1, (_T("+SensorSetup\n")));
	SensorSetup();
	//DEBUGMSG(1, (_T("-SensorSetup\n")));
    
    g_SensorI2C_Addr = I2C_ADDRESS;
    g_SensorI2C_Speed = I2C_400KBPS;
    g_SensorI2C_OffsetWidth = 2;
    g_SensorI2C_DataWidth = 1;

	DEBUGMSG(ZONE_FUNC, (_T("-SensorYuv_Open OK\n")));
    return NV_TRUE;    

fail:
    free(pSensorYuvContext);
    return NV_FALSE;
}

/**
 * SensorYuv_Close. Phase 1.
 * Free the sensor's private context
 */
static void SensorYuv_Close(NvOdmImagerHandle hImager)
{
    SensorYuvContext *pContext = NULL;
        
    if (!hImager || !hImager->pSensor || !hImager->pSensor->pPrivateContext)
        return;

    pContext = (SensorYuvContext*)hImager->pSensor->pPrivateContext;

    // cleanup
    free(pContext);
    hImager->pSensor->pPrivateContext = NULL;
}

/**
 * SensorYuv_GetCapabilities. Phase 1.
 * Returnt sensor Yuv's capabilities
 */
static void
SensorYuv_GetCapabilities(
    NvOdmImagerHandle hImager,
    NvOdmImagerCapabilities *pCapabilities)
{
	DEBUGMSG(ZONE_FUNC, (L"+SensorYuv_GetCapabilities\n"));
    // Copy the sensor caps from g_SensorYuvCaps
    memcpy(pCapabilities, &g_SensorYuvCaps,
        sizeof(NvOdmImagerCapabilities));

	DEBUGMSG(ZONE_FUNC, (L"-SensorYuv_GetCapabilities\n"));
}

/**
 * SensorYuv_ListModes. Phase 1.
 * Return a list of modes that sensor yuv supports.
 * If called with a NULL ptr to pModes, then it just returns the count
 */
static void
SensorYuv_ListModes(
    NvOdmImagerHandle hImager,
    NvOdmImagerSensorMode *pModes,
    NvS32 *pNumberOfModes)
{
    SensorYuvContext *pContext =
        (SensorYuvContext*)hImager->pSensor->pPrivateContext;
    NvS32 i;

	DEBUGMSG(ZONE_FUNC, (L"+SensorYuv_ListModes\n"));

    if (pNumberOfModes)
    {
        if(*pNumberOfModes!=-1 && (i=GetSensorMode())!=-1)
        {
            *pNumberOfModes=1;
            if (pModes)
                pModes[0] = g_SensorYuvSetModeSequenceList[i].Mode;
        }
        else
        {
            *pNumberOfModes = (NvS32)pContext->NumModes;
            if (pModes)
            {
                // Copy the modes from g_SensorYuvSetModeSequenceList
                for (i = 0; i < *pNumberOfModes; i++)
                    pModes[i] = g_SensorYuvSetModeSequenceList[i].Mode;
            }
        }
    }

	DEBUGMSG(ZONE_FUNC, (L"-SensorYuv_ListModes\n"));
}

static BOOL bFirst = FALSE;

/**
 * SensorYuv_SetMode. Phase 1.
 * Set sensor yuv to the mode of the desired resolution and frame rate.
 */
static NvBool
SensorYuv_SetMode(
    NvOdmImagerHandle hImager,
    const SetModeParameters *pParameters,
    NvOdmImagerSensorMode *pSelectedMode,
    SetModeParameters *pResult)
{
    NvBool Status = NV_FALSE;

    SensorYuvContext *pContext = (SensorYuvContext*)hImager->pSensor->pPrivateContext;

    NvU32 Index;

	DEBUGMSG(ZONE_FUNC, (_T("+SensorYuv_SetMode %dx%d\n"), pParameters->Resolution.width, pParameters->Resolution.height));

    //ASSERT(pParameters);

    NKDbgPrintfW(L"Setting resolution to %dx%d\n",
        pParameters->Resolution.width, pParameters->Resolution.height);

    // Find the right entrys in g_SensorYuvSetModeSequenceList that matches
    // the desired resolution and framerate
    for (Index = 0; Index < pContext->NumModes; Index++)
    {
        if ((pParameters->Resolution.width == g_SensorYuvSetModeSequenceList[Index].Mode.ActiveDimensions.width) &&
            (pParameters->Resolution.height == g_SensorYuvSetModeSequenceList[Index].Mode.ActiveDimensions.height) /*&&
            (pContext->FrameRate == g_SensorYuvSetModeSequenceList[Index].Mode.PeakFrameRate)*/ )
            break;
    }

    // No match found
	if (Index == pContext->NumModes) {
		DEBUGMSG(ZONE_ERR, (_T("-SensorYuv_SetMode no match found\n")));
        return NV_FALSE;
	}

    if (pSelectedMode)
    {
		DEBUGMSG(ZONE_INFO, (_T("Setting resolution: %dx%d with %d fps.\n"), g_SensorYuvSetModeSequenceList[Index].Mode.ActiveDimensions.width, g_SensorYuvSetModeSequenceList[Index].Mode.ActiveDimensions.height, (NvU32)g_SensorYuvSetModeSequenceList[Index].Mode.PeakFrameRate));
        *pSelectedMode = g_SensorYuvSetModeSequenceList[Index].Mode;
    }

	if (pContext->ModeIndex == Index) {
		DEBUGMSG(ZONE_FUNC, (_T("-SensorYuv_SetMode same mode\n")));
		return NV_TRUE;
	}

	//if(!SensorSetup()) {
	//	DEBUGMSG(1, (_T("-SensorYuv_SetMode OV5642 SetupUp mode change failed\n")));
	//	return NV_FALSE;
	//}
	
	Sleep(250);

    // i2c writes for the set mode sequence
	DEBUGMSG(ZONE_INFO, (_T(" SensorYuv_SetMode %d Write ModeSequence\n"), Index));
    Status = SensorI2C_WriteSequence(g_SensorYuvSetModeSequenceList[Index].pSequence);
    if(bFirst == FALSE)
    {				
        DEBUGMSG(ZONE_INFO, (_T("ov5640_af_reg Set\n"), Index));
        SensorI2C_WriteSequence(ov5640_af_reg);      
        bFirst  = TRUE;
    } 

	if (!Status) {
		DEBUGMSG(ZONE_FUNC, (_T("-SensorYuv_SetMode OV5642 I2C failed\n")));
       return NV_FALSE;
	}

    pContext->ModeIndex = Index;

    if (pResult)
    {
        pResult->Resolution = g_SensorYuvSetModeSequenceList[Index].
                              Mode.ActiveDimensions;
        pResult->Exposure = 0;
        memset(pResult->Gains, 0, sizeof(NvF32) * 4);
    }

	DEBUGMSG(ZONE_FUNC, (_T("-SensorYuv_SetMode done\n")));
    return Status;
}

/**
 * SensorYuv_SetPowerLevel. Phase 1
 * Set the sensor's power level
 */
static NvBool
SensorYuv_SetPowerLevel(
    NvOdmImagerHandle hImager,
    NvOdmImagerPowerLevel PowerLevel)
{
    NvBool Status = NV_TRUE;
    SensorYuvContext *pContext =
        (SensorYuvContext*)hImager->pSensor->pPrivateContext;
    //const NvOdmPeripheralConnectivity *pConnections;
    const char* powerlevel_str[6] = {"?0", "Off","StandBy", "On", "SyncSensors"};

	DEBUGMSG(ZONE_FUNC, (_T("+SensorYuv_SetPowerLevel %S\n"), powerlevel_str[PowerLevel]));

    if (pContext->PowerLevel == PowerLevel)
        goto exit;

    ////NV_ASSERT(hImager->pSensor->pConnections);
    //pConnections = hImager->pSensor->pConnections;

    switch(PowerLevel)
    {
        case NvOdmImagerPowerLevel_On:
        SensorSetup();
            break;

        case NvOdmImagerPowerLevel_Standby:
/*
            Status = SetPowerLevelWithPeripheralConnectivity(pConnections,
                         &pContext->I2c, &pContext->GpioConfig,
                         NvOdmImagerPowerLevel_Standby);
*/
            break;
        case NvOdmImagerPowerLevel_Off:
/*
            Status = SetPowerLevelWithPeripheralConnectivity(pConnections,
                         &pContext->I2c, &pContext->GpioConfig,
                         NvOdmImagerPowerLevel_Off);
            
            // wait for the standby 0.1 sec should be enough
            NvOdmOsWaitUS(100000); 
*/
            break;
        default:
//            //NV_ASSERT("!Unknown power level\n");
            Status = NV_FALSE;
            break;
    }

    if (Status)
        pContext->PowerLevel = PowerLevel;

exit:
	DEBUGMSG(ZONE_FUNC, (L"-SensorYuv_SetPowerLevel returns %s\n", Status?L"True":L"False"));

    return Status;
}




/**
 * SensorYuv_SetParameter. Phase 2.
 * Set sensor specific parameters. 
 * For Phase 1. This can return NV_TRUE.
 */
static NvBool
SensorYuv_SetParameter(
    NvOdmImagerHandle hImager,
    NvOdmImagerParameter Param,
    NvS32 SizeOfValue,
    const void *pValue)
{
    NvBool Status = NV_TRUE;

    BOOL bDebugEnable = TRUE;

#if DEBUG //NvOdmImagerParameter_SensorFrameRate is called too often, output it only once
    {
        static NvOdmImagerParameter OldParam = 0;

        if(Param==NvOdmImagerParameter_SensorTriggerStill)
        {
            if(OldParam&1)
                bDebugEnable=FALSE;
            OldParam |= 1;
        }
        /*if(Param==xxd)
        {
            if(OldParam&2)
                bDebugEnable=FALSE;
            OldParam |= 2;
        }*/
    }
#endif

    DEBUGMSG(ZONE_FUNC && bDebugEnable, (L"+SensorYuv_SetParameter(%S)\n", GetParamName(Param)));
    
    switch(Param)
    {
        case NvOdmImagerParameter_SensorTriggerStill:
        {
            static NvU32 t0=-1;
            NvU32 t1;
            static NvU32 frames=0;

            if(t0==-1)
                t0=GetTickCount();

            frames++;
            t1 = GetTickCount();

            if(t1-t0>=1000)
            {
                DEBUGMSG(ZONE_INFO, (_T("Average fps: %d\n"), frames));
                t0 = t1;
                frames=0;
            }
            
            break;
        }
        case NvOdmImagerParameter_SensorInputClock:

            if(SizeOfValue == sizeof(NvOdmImagerClockProfile))
            {
                // Verify the input clock is valid.
                NvU32 InClock =
                    ((NvOdmImagerClockProfile*)pValue)->ExternalClockKHz;

                DEBUGMSG(ZONE_FUNC && ZONE_INFO, (L"Clock=%d KHz\r\n", InClock));
				
				if (!InClock) {
					return NV_FALSE;
				}
                //if (InClock < g_SensorYuvCaps.InitialSensorClockRateKHz)
                //    return NV_FALSE;
            }
            break;

        case NvOdmImagerParameter_OptimizeResolutionChange:
			if(SizeOfValue!=sizeof(NvBool)){break;}

            // Sensor yuv doesn't support optimized resolution change.
            //if (*((NvBool*)pValue) == NV_FALSE)
			if (*((NvBool*)pValue) == NV_TRUE)
                return NV_TRUE;
            else
                return NV_FALSE;

            break;

        case NvOdmImagerParameter_SelfTest:
            // not implemented.
            break;
        default:
            DEBUGMSG(ZONE_FUNC && ZONE_WARN && bDebugEnable, (L"Not Implemented\n"));
            Status = NV_FALSE;
            break;
    }


	DEBUGMSG(ZONE_FUNC && bDebugEnable, (L"-SensorYuv_SetParameter\n"));

    return Status;
}

/**
 * SensorYuv_GetParameter. Phase 1.
 * Get sensor specific parameters
 */
static NvBool
SensorYuv_GetParameter(
    NvOdmImagerHandle hImager,
    NvOdmImagerParameter Param,
    NvS32 SizeOfValue,
    void *pValue)
{
    NvBool result = NV_TRUE;

    SensorYuvContext *pContext =
        (SensorYuvContext*)hImager->pSensor->pPrivateContext;
#if DEBUG //NvOdmImagerParameter_SensorFrameRate is called too often, output it only once
    {
        BOOL bDebugEnable = TRUE;
        static NvOdmImagerParameter OldParam = -1;

        if(Param==NvOdmImagerParameter_SensorFrameRate &&
            OldParam==NvOdmImagerParameter_SensorFrameRate )
            bDebugEnable=FALSE;
        OldParam = Param;
        if(!bDebugEnable)
            return FALSE;
    }
#endif

    DEBUGMSG(ZONE_FUNC, (L"+SensorYuv_GetParameter(%S)\n", GetParamName(Param)));

    switch(Param)
    {
        case NvOdmImagerParameter_DeviceStatus:
            if(SizeOfValue ==sizeof(NvOdmImagerDeviceStatus))
            {
                NvOdmImagerDeviceStatus *pStatus;
                pStatus = (NvOdmImagerDeviceStatus *)pValue;
/*
                {
                    NvOdmImagerI2cConnection *pI2c = &pContext->I2c;
                    // Pick your own useful registers to use for debug.
                    // If the camera hangs, these register values are printed
                    // Sensor Dependent.
                    NvOdmImagerI2cWrite(pI2c, 0x098C, 0x0000);
                    NvOdmImagerI2cRead(pI2c, 0x0990, &pStatus->Values[0]);
                    NvOdmImagerI2cWrite(pI2c, 0x098C, 0x2104);
                    NvOdmImagerI2cRead(pI2c, 0x0990, &pStatus->Values[1]);
                    NvOdmImagerI2cWrite(pI2c, 0x098C, 0x2703);
                    NvOdmImagerI2cRead(pI2c, 0x0990, &pStatus->Values[2]);
                    NvOdmImagerI2cWrite(pI2c, 0x098C, 0x2705);
                    NvOdmImagerI2cRead(pI2c, 0x0990, &pStatus->Values[3]);
                    NvOdmImagerI2cWrite(pI2c, 0x098C, 0x2737);
                    NvOdmImagerI2cRead(pI2c, 0x0990, &pStatus->Values[4]);

                    pStatus->Count = 5;
                }
*/
            }
            break;

        case NvOdmImagerParameter_SensorFrameRate:
			if(SizeOfValue == sizeof(NvF32)){            
                *((NvF32*)pValue) = g_SensorYuvSetModeSequenceList[pContext->ModeIndex].Mode.PeakFrameRate;
                break;
			}

        case NvOdmImagerParameter_GetBestSensorMode:
            if(SizeOfValue == sizeof(NvOdmImagerSensorMode))
            {
                NvOdmImagerSensorMode* pMode = (NvOdmImagerSensorMode *) pValue;
                DEBUGMSG(ZONE_INFO, (L"ModeToSet(%d): %dx%d (%d fps)\r\n", pMode->Type, pMode->ActiveDimensions.width, pMode->ActiveDimensions.height, (NvU32)pMode->PeakFrameRate));
                pContext->FrameRate = pMode->PeakFrameRate;
                //return NV_TRUE;
                result = NV_FALSE;
                break;
            }

        case NvOdmImagerParameter_ISPSetting:
            if (pValue &&
                SizeOfValue == sizeof(NvOdmImagerISPSetting))
            {
                NvOdmImagerISPSetting *pSetting =
                    (NvOdmImagerISPSetting *)pValue;

                DEBUGMSG(ZONE_INFO, (L"ISP Setting: 0x%x %d\r\n",pSetting->attribute, pSetting->dataSize));

                switch(pSetting->attribute)
                {
                case NvCameraIspAttribute_FlickerFrequency:
                    break;
                }
            }
            break;
        default:
           //NV_ASSERT(!"Not Implemented");
           DEBUGMSG(ZONE_FUNC && ZONE_WARN, (L"Not Implemented\n"));
           result = NV_FALSE;
           break;
    }

	DEBUGMSG(ZONE_FUNC, (L"-SensorYuv_GetParameter\n"));

    return result;
}

/**
 * SensorYuv_GetPowerLevel. Phase 2.
 * Get the sensor's current power level
 */
static void
SensorYuv_GetPowerLevel(
    NvOdmImagerHandle hImager,
    NvOdmImagerPowerLevel *pPowerLevel)
{
	

    SensorYuvContext *pContext =
        (SensorYuvContext*)hImager->pSensor->pPrivateContext;


	DEBUGMSG(ZONE_FUNC, (L"+SensorYuv_GetPowerLevel\n"));

    *pPowerLevel = pContext->PowerLevel;

	DEBUGMSG(ZONE_FUNC, (L"-SensorYuv_GetPowerLevel\n"));
}

/**
 * SensorYuv_GetHal. Phase 1.
 * return the hal functions associated with sensor yuv
 */
NvBool SensorPlugin_GetHal_YUV_OV5640_Parallel(NvOdmImagerHandle hImager)
{
    if (!hImager || !hImager->pSensor)
        return NV_FALSE;

    hImager->pSensor->pfnOpen = SensorYuv_Open;
    hImager->pSensor->pfnClose = SensorYuv_Close;
    hImager->pSensor->pfnGetCapabilities = SensorYuv_GetCapabilities;
    hImager->pSensor->pfnListModes = SensorYuv_ListModes;
    hImager->pSensor->pfnSetMode = SensorYuv_SetMode;
    hImager->pSensor->pfnSetPowerLevel = SensorYuv_SetPowerLevel;
    hImager->pSensor->pfnGetPowerLevel = SensorYuv_GetPowerLevel;
    hImager->pSensor->pfnSetParameter = SensorYuv_SetParameter;
    hImager->pSensor->pfnGetParameter = SensorYuv_GetParameter;

    return NV_TRUE;
}
